from __future__ import annotations

import abc
import asyncio
import functools
import logging
import types
import warnings
from typing import (
    Awaitable,
    Generic,
    Coroutine,
    Mapping,
    Any,
    Callable,
    Tuple,
)

from . import util
from .util.typing import (
    T_EnumFlag,
    Descriptor,
)

Callback = Callable[..., Coroutine[Any, Any, None]]
_StateChange = Tuple[T_EnumFlag, T_EnumFlag]

log = logging.getLogger(__name__)


class AbstractProtocol(Generic[T_EnumFlag], abc.ABC):
    """
    ABC to implement a state machine with async callbacks. A :class:`AbstractProtocol` can be used as an async context
    manager, to `start` and `stop` the protocol automatically, or be "manually" started and stopped using the protocols
    methods.
    """

    state_changes: Mapping[Tuple[T_EnumFlag | None, ...], Callback]
    """
    Assign to this mapping in your concrete implementation to define callbacks for state changes.
    """
    starting_state: T_EnumFlag
    """
    Assign a state to this attribute as the protocols starting state when implementing a concrete protocol
    """
    end_state: T_EnumFlag
    """
    Assign a state to this attribute as the protocols end state when implementing a concrete protocol
    """

    @property
    def finished(self) -> bool:
        """
        Convenience property that simply returns whether the protocols state is its end state.
        """
        return self.state.value == self.end_state

    @property
    def was_started(self) -> bool:
        """
        Convenience property that returns whether a protocol has already been started
        """
        return self._run is not None

    @util.cached_property
    def context(self) -> types.SimpleNamespace:
        """
        It's encouraged to overwrite this and return a dataclass or something that is better type-able  in your
        concrete implementation
        """
        return types.SimpleNamespace()

    def __init__(self) -> None:
        super().__init__()
        self.__name__: str = self.__class__.__name__
        self._state = None
        self._run: asyncio.Task | None = None
        self.change_context = asyncio.Condition()
        """
        Condition to notify / wait for changed :attr:`.context`
        """
        self.task_nursery = util.TaskNursery(name=lambda: f"Task Nursery for {self}")
        """
        This manages the tasks created by the protocol as well as teardown behaviour if the protocol is stopped
        """
        self.get_state_change_callback = functools.partial(util.enum.EnumMatcher.get_matching_value,
                                                           mapping=self.state_changes)
        """
        This callable uses :meth:`~ubii.framework.util.EnumMatcher.get_matching_value`
        to return the appropriate state change from :attr:`.state_changes` for a tuple of states
        """

    def _get_state(self) -> T_EnumFlag:
        return self._state

    def _set_state(self, new_state: T_EnumFlag):
        current = self._state
        if new_state == current:
            return

        # it is allowed to change the state if a matching callback is defined
        if not self.get_state_change_callback((current, new_state), None):
            raise ValueError(f"Can't change state {current!r} -> {new_state!r}")

        self._state = new_state

    # declaring the property this way helps pycharm to infer types correctly see
    # https://youtrack.jetbrains.com/issue/PY-15176 and related issues.
    state: util.condition_property = util.condition_property(fget=_get_state, fset=_set_state, )
    """
    Makes shared access to the protocol state easy
    """

    def start(self: AbstractProtocol[T_EnumFlag]) -> AbstractProtocol[T_EnumFlag]:
        """
        Starts the protocol task.

        It is encouraged to wait for the protocol if the task is cancelled instead of waiting for the task.
        In either case, canceling the task will trigger the sentinel task to handle the
        protocol teardown, which might - depending on the protocol implementation - mean that closing the event loop
        shortly after the cancellation might raise exceptions from the scheduled cleanup operations, so it's
        mandatory to sleep for a short time (e.g. asyncio.sleep(0)) before closing the loop if your protocol
        schedules async tasks during its teardown.

        Returns:
            reference to `self` -- the started protocol

        See Also:
            :class:`RunProtocol` -- specialized coroutine that is used to create the task

            :attr:`.task_nursery` -- the object that handles cancellation / stopping of the task when
            necessary
        """
        if self.was_started:
            warnings.warn(f"{self} already running.")
            return self

        async def stop_if_running(protocol: AbstractProtocol):
            if protocol.was_started and not protocol.finished:
                await protocol.stop()

        self._run = self.task_nursery.create_task(RunProtocol(self))
        self.task_nursery.push_async_callback(lambda: stop_if_running(self))
        return self

    async def stop(self) -> None:
        """
        Gracefully shut down the protocol by setting the protocol state to the end state.
        This will call the appropriate state change callback (for the :math:`current \\rightarror end` state change)
        and finish the task created by :meth:`.start`

        Returns control back to the caller after the task created by :meth:`.start` stopped and all teardown
        callbacks have been scheduled
        """
        assert self.was_started, "Protocol not running, `start()` first"
        await self.state.set(self.end_state)
        await self._run
        await self.task_nursery.aclose()
        self._run = None

    async def __aenter__(self):
        self.start()
        await self.state.get(predicate=lambda value: value == self.starting_state)
        return self

    def __aexit__(self, *exc_info):
        return self.task_nursery.__aexit__(*exc_info)


@util.dunder.repr('protocol')
class RunProtocol(util.CoroutineWrapper):
    """
    A specialized coroutine to wait for changing protocol :attr:`~AbstractProtocol.state`, get the appropriate callbacks
    from the protocols :attr:`~AbstractProtocol.state_changes` and execute them.

    Starting a protocol using the :meth:`~AbstractProtocol.start` method, starts a :class:`RunProtocol` coroutine
    as a task using the protocol's :attr:`AbstractProtocol.task_nursery`.
    """

    def __init__(self, protocol: AbstractProtocol):
        """
        Args:
            protocol: a protocol to run
        """
        self.protocol = protocol
        """
        Reference to the protocol that is running in this coroutine
        """
        self.__name__ = repr(self.protocol)
        super().__init__(coroutine=self._run())

    @staticmethod
    async def _no_callback_found(protocol, context):
        raise RuntimeError(f"No callback found for state_change {context.state_change} in {protocol}")

    async def _run_state_change_callback(self, prev, cur, ctx):
        cb = self.protocol.get_state_change_callback((prev, cur), self._no_callback_found)
        ctx.state_change = (prev, cur)
        log.debug(f"Changed state {prev!r}->{cur!r} in {self.protocol}, got callback {getattr(cb, '__name__', cb)!r}")

        # all functions are also descriptors
        assert isinstance(cb, Descriptor), f"{cb} needs to be a descriptor, e.g. a function"
        coro = cb.__get__(self.protocol, type(self.protocol))(ctx)

        if not isinstance(coro, Awaitable):
            raise RuntimeError(f"{cb} is not awaitable (not using `async def`?)")

        return await coro

    async def _run(self):
        await self.protocol.state.set(self.protocol.starting_state)
        previous = None
        current = self.protocol.state.value
        end_state = self.protocol.end_state
        context = self.protocol.context

        while previous != end_state:
            async with self.protocol.change_context:
                try:
                    await self._run_state_change_callback(previous, current, context)
                    self.protocol.change_context.notify_all()
                except Exception as initial:
                    if self.protocol.state.value == current:
                        raise
                    else:
                        previous = current
                        current = self.protocol.state.value

                        log.debug(f"Changed state during exception {initial}: {previous!r} -> {current!r}")
                        try:
                            result = await self._run_state_change_callback(previous, current, context)
                        except Exception as nested:
                            raise nested from initial
                        else:
                            if not result:
                                raise initial

            previous = current
            if current != end_state:
                # wait until a state is set that is not the current state
                current = await self.protocol.state.get(
                    predicate=lambda value: value != current
                )

        log.debug(f"{self} finished.")



from __future__ import annotations

import abc
import fnmatch
import typing
from operator import add

from .functools import hook, document_decorator
from .typing import T_co


def merge_dicts(base: typing.Dict, merge: typing.Dict, op: typing.Callable = add) -> typing.Dict:
    def merge_op(left, right):
        if isinstance(left, dict) and isinstance(right, dict):
            return merge_dicts(left, right, op=op)
        else:
            return op(left, right)

    return {**base, **merge, **{k: merge_op(base[k], merge[k]) for k in base if k in merge}}


class MatchMappingMixin(typing.Mapping[str, T_co], abc.ABC):
    """
    Mixin that implements glob pattern matches on mapping keys
    """

    def match_name(self, name) -> typing.Tuple[T_co, ...]:
        """
        Returns all values where ``name`` matches the keys of contained values interpreted as a glob pattern.

        Example:

            >>> import collections
            >>> class Container(collections.OrderedDict, MatchMappingMixin): pass
            >>> container = Container({"foo": 1, "foo*": 2, "bar": 3})
            >>> val = container.match_name('foo')
            >>> print(val)
            (1, 2)

        See Also:
            :mod:`fnmatch` -- details about glob patterns

        """
        return tuple(val for pattern, val in self.items() if fnmatch.fnmatch(name=name, pat=pattern))

    def match_pattern(self, pattern) -> typing.Tuple[T_co, ...]:
        """
        Returns all values where the keys of contained values match the glob ``pattern``.

        Example:

            >>> import collections
            >>> class Container(collections.OrderedDict, MatchMappingMixin): pass
            >>> container = MatchMappingMixin({"foo": 1, "foo*": 2, "bar": 3})
            >>> val = container.match_pattern('foo')
            >>> print(val)
            (1)

        See Also:
            :mod:`fnmatch` -- details about glob patterns

        """
        return tuple(top for topic_pattern, top in self.items() if fnmatch.fnmatch(name=topic_pattern, pat=pattern))


T_Key = typing.TypeVar('T_Key')


class DefaultHookMap(typing.Generic[T_Key, T_co], typing.Mapping[T_Key, T_co]):
    """
    Acts like a :class:`~collections.defaultdict` mapping but it's :attr:`.default_factory` is always a
    :class:`util.hook` so it can be adjusted after creation of the mapping
    """

    def __init__(self: DefaultHookMap[T_Key, T_co], base_factory: typing.Callable[[T_Key], T_co]):
        """
        Args:
            base_factory: Saved as :attr:`.base_factory`
        """
        self.base_factory: typing.Callable[[T_Key], T_co] = base_factory
        """
        The base factory callable, gets used in :func:`.default_factory` to actually create the item
        """
        self.data: typing.Dict[str, T_co] = {}

    @hook
    @document_decorator(hook)
    def default_factory(self, key: T_Key) -> None:
        """
        Called whenever a key is not present in the mapping.
        Creates new value using the :attr:`.base_factory`

        Example:

            >>> from ubii.framework.util import DefaultHookMap
            >>> class Topic:
            ...     def __init__(self, pattern):
            ...             self.pattern = pattern
            ...
            >>> store = DefaultHookMap(default_factory=Topic)
            >>> topic = store['topic/glob/pattern']
            >>> assert topic.pattern == 'topic/glob/pattern'

        Args:
            key: glob pattern

        """
        self.data[key] = self.base_factory(key)

    def __getitem__(self, key: T_Key) -> T_co:
        if key not in self.data:
            self.default_factory(key)
        assert key in self.data
        return self.data[key]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> typing.Iterator[T_Key]:
        return iter(self.data)

    def __contains__(self, item: T_Key):
        return item in self.data

    def __delitem__(self, key: T_Key):
        if key not in self.data:
            raise KeyError(f"Can't delete item with key {key}")

        del self.data[key]



import asyncio
import logging
import types
import uuid

import pytest
import ubii.proto as ub

from ubii.framework import processing
from ubii.framework.client import RunProcessingModules, Subscriptions, Publish, Sessions

pytestmark = pytest.mark.asyncio

__protobuf__ = ub.__protobuf__

log = logging.getLogger(__name__)

MODULE_NAME = "example-processing-module"
PARAM_NAME = 'test'

js_on_processing_stringified = ub.ProcessingModule(
    on_processing_stringified='\n'.join((
        "function processingCallback(deltaTime, inputs) {",
        "  let outputs = {};",
        f"  outputs.{PARAM_NAME} = !inputs.{PARAM_NAME};",
        "  return {outputs};",
        "}",
    )),
    language=ub.ProcessingModule.Language.JS,
)

py_on_processing_stringified = ub.ProcessingModule(
    on_processing_stringified='\n'.join((
        'def on_processing(self, context):',
        '    import logging',
        '    import ubii.proto',
        '    log = logging.getLogger("ubii.interact.protocol")',
        '    log.info(f"------- delta_time: {context.delta_time} ---------")',
        f'    input_record = context.inputs.{PARAM_NAME}',
        f'    context.outputs.{PARAM_NAME} = ubii.proto.TopicDataRecord(bool=not input_record.bool)',
    )),
    language=ub.ProcessingModule.Language.PY,
)

py_on_created_stringified = ub.ProcessingModule(
    on_created_stringified='\n'.join((
        'def on_created(self, context):',
        '    import logging',
        '    log = logging.getLogger("ubii.interact.protocol")',
        '    log.info(f"------- TEST: {context} ---------")',
        '',
    )),
    language=ub.ProcessingModule.Language.PY,
)


class Processing:
    @pytest.fixture(scope='class')
    def base_module(self):
        processing_module = ub.ProcessingModule(
            name=MODULE_NAME,
            processing_mode={'trigger_on_input': {'min_delay_ms': 0,
                                                  'all_inputs_need_update': True}},

            on_processing_stringified="",
            inputs=[
                {
                    'internal_name': PARAM_NAME,
                    'message_format': 'bool'
                },
            ],
            outputs=[
                {
                    'internal_name': PARAM_NAME,
                    'message_format': 'bool'
                }
            ],
        )
        return processing_module

    @pytest.fixture(scope='class')
    async def base_session(self, client) -> ub.Session:
        await client
        module = client.processing_modules[0]
        input_topic = f"{client.id}/test_input"
        output_topic = f"{client.id}/test_output"
        io_mappings = [
            {
                'processing_module_name': module.name,
                'input_mappings': [
                    {
                        'input_name': module.inputs[0].internal_name,
                        'topic': input_topic
                    },
                ],
                'output_mappings': [
                    {
                        'output_name': module.outputs[0].internal_name,
                        'topic': output_topic
                    }
                ]
            },
        ]

        session = ub.Session(name=str(uuid.uuid4()),
                             processing_modules=[module],
                             io_mappings=io_mappings)

        yield session

    @pytest.fixture
    def pm_startup(self):
        pass

    @pytest.fixture
    async def test_value(self, pm_startup, client, base_session: ub.Session):
        topic, = await client[Subscriptions].subscribe_topic(base_session.io_mappings[0].output_mappings[0].topic)
        yield topic.buffer

    @pytest.mark.parametrize('data', [False, True])
    @pytest.mark.parametrize('delay', [1, 0.5, 0.2, 0.1])
    @pytest.mark.parametrize('timeout', [0.4])
    async def test_processing_module(self, client, test_value, base_session: ub.Session, delay, timeout, data):
        await asyncio.sleep(delay)
        await client[Publish].publish({'topic': base_session.io_mappings[0].input_mappings[0].topic, 'bool': data})
        record = await asyncio.wait_for(test_value.get(), timeout=timeout)

        assert isinstance(record.bool, bool)
        assert record.bool != data


class TestPy(Processing):
    module_spec = [
        pytest.param((py_on_processing_stringified, py_on_created_stringified), id='python')
    ]

    client_spec = [
        pytest.param((ub.Client(
            is_dedicated_processing_node=True,
            processing_modules=[ub.ProcessingModule(name=MODULE_NAME)]
        ),), id='processing_node')
    ]

    @pytest.fixture
    async def pm_startup(self, client, module_spec, base_session):
        await client.implements(RunProcessingModules)
        await client.implements(Sessions)
        started = await client[Sessions].start_session(base_session)

        pm = await client[RunProcessingModules].get_module_instance(
            module_spec.name, module_spec.Status.CREATED
        )
        yield pm

        await client[Sessions].stop_session(started)


@pytest.mark.xfail(reason="Broker bug")
class TestJS(Processing):
    module_spec = [
        pytest.param((js_on_processing_stringified,), id='js')
    ]

    client_spec = [
        pytest.param((ub.Client(is_dedicated_processing_node=False, processing_modules=None),), id='server_processing')
    ]


class TestLateInitModules(TestPy):
    class FakeModule(processing.ProcessingRoutine):
        """
        Test Module
        """

        def __init__(self, context, **kwargs):
            super().__init__(**kwargs)
            self.name = MODULE_NAME
            constants: ub.Constants = context.constants
            self.tags = ['pytest']
            self.description = 'Test Module'

            self.inputs = [
                {
                    'internal_name': 'test',
                    'message_format': constants.MSG_TYPES.DATASTRUCTURE_BOOL
                },
            ]

            self.outputs = [
                {
                    'internal_name': 'test',
                    'message_format': constants.MSG_TYPES.DATASTRUCTURE_BOOL
                },
            ]

            self.processing_mode = {
                'frequency': {
                    'hertz': 20
                }
            }
            self._context = context

        def on_processing(self, context: types.SimpleNamespace) -> None:
            input_value: ub.TopicDataRecord = getattr(context.inputs, PARAM_NAME)
            if input_value:
                setattr(context.outputs, PARAM_NAME, ub.TopicDataRecord(bool=not input_value.bool))

    late_init_module_spec = [
        pytest.param(FakeModule, id=MODULE_NAME)
    ]

    module_spec = []

    @pytest.mark.parametrize('data', [False, True])
    @pytest.mark.parametrize('timeout', [0.4])
    @pytest.mark.parametrize('run_time', [4])
    @pytest.mark.parametrize('allowed_relative_error', [0.9])
    async def test_processing_module(self,
                                     client,
                                     data,
                                     base_session: ub.Session,
                                     timeout,
                                     run_time,
                                     pm_startup,
                                     allowed_relative_error):
        received = []
        input_topic = base_session.io_mappings[0].input_mappings[0].topic
        output_topic = base_session.io_mappings[0].output_mappings[0].topic

        await client[Publish].publish({'topic': input_topic, 'bool': data})
        await asyncio.sleep(0.2)  # we wait until the broker processed the publishing
        topics, tokens = await client[Subscriptions].subscribe_topic(output_topic).with_callback(received.append)
        # if you subscribe to the same topic as before, you will immediately get the last value, therefore
        # we publish before we subscribe!

        await asyncio.sleep(run_time)
        await topics[0].unregister_callback(tokens[0], timeout=timeout)
        await client[Subscriptions].unsubscribe_topic(output_topic)

        assert all('bool' in r and r.bool != data for r in received), received
        tested_module: ub.ProcessingModule = base_session.processing_modules[0]

        max_computations = tested_module.processing_mode.frequency.hertz * run_time
        assert len(received) / max_computations >= allowed_relative_error



from __future__ import annotations

import asyncio
import collections
import concurrent.futures
import contextlib
import dataclasses
import enum
import functools
import logging
import types
import typing
import warnings

import sys
import ubii.proto

import codestare.async_utils
from . import (
    protocol,
    util,
    topics,
    debug
)
from .util.typing import AsyncGetter, AsyncSetter

__protobuf__ = ubii.proto.__protobuf__

log = logging.Logger(__name__)

__json_name_to_field_name__ = {field.json_name: field.name for field in
                               ubii.proto.TopicDataRecord.pb().DESCRIPTOR.fields}
"""
Used in :meth:`.ProcessingProtocol.helpers.fix_io_fmt` to compute the field name of the 
:attr:`ubii.proto.TopicDataRecord.type` one-of group from the message type given by a 
:attr:`ubii.proto.ModuleIO.message_format` field. 
"""


def fix_io_fmt(message_format: str) -> str:
    """
    Computes the field name of the :attr:`~ubii.proto.TopicDataRecord.type` oneof corresponding to the
    type in a :attr:`~ubii.proto.ModuleIO.message_format` field (of the form ``ubii.{proto_package}.{type}``
    as defined in the .proto file, not the python package!)

    Example:

        The :attr:`~ProcessingRoutine.inputs` of a routine contain a
        :class:`ubii.proto.ModuleIO` message with :attr:`~ubii.proto.ModuleIO.message_format`
        ``ubii.dataStructure.Matrix4x4``. The name of the corresponding field in a
        :class:`ubii.proto.TopicDataRecord` is ``matrix4x4``.
        This method performs this conversion.

    Args:
        message_format: format string for message type

    Returns:
        name of corresponding field in a :class:`ubii.proto.TopicDataRecord` message

    """
    if not message_format.startswith('ubii.'):
        return message_format

    ubii_name = ubii.proto.util.get_import_name(message_format)
    field_name = __json_name_to_field_name__.get(
        f'{ubii_name.type[0].lower() + ubii_name.type[1:]}'
    )
    assert field_name, f"fixing field name for {message_format} failed."
    return field_name


MAX_WORKERS = 8


class perf_calc_callable(util.typing.Protocol):
    def __call__(self, scheduler: Scheduler) -> float:
        """
        This should evalutate the performance of a scheduler instance somehow.

        Args:
            scheduler: scheduler instance

        Returns:
            a value indicating the performance
        """


def perf_calc(attr: str, scheduler: Scheduler) -> float:
    """
    Args:
        scheduler: scheduler instance
        attr: name of attribute of the scheduler which contains the data for the performance evaluation.

    Returns:
        performace of scheduler, calculated by computing relative error of data to
        :attr:`~Scheduler.delay` i.e. for an average of past execution times
        :math:`\\overline{t} = avg(\\text{scheduler.exec_delta_times})` return :math:`1` if
        :math:`\\overline{t} < \\text{scheduler.delay}` otherwise calculate the relative deviation
        :math:`\\Delta t = \\frac{\\overline{t} - \\text{scheduler.delay}}{\\text{scheduler.delay}}` and return
        :math:`1 - \\Delta t` (negative values if :math:`\\Delta t > 1` allowed)

    """
    if 'lockstep' in scheduler.mode:
        warnings.warn(f"This performance rating is not applicable to schedulers in lockstep mode")

    values = getattr(scheduler, attr, ())
    if not values:
        return 1.

    avg = sum(values) / len(values)
    return 1. - (avg - scheduler.delay) / scheduler.delay  # 1 - rel error


def check_datatype(io: ubii.proto.ModuleIO, record: ubii.proto.TopicDataRecord):
    """
    Check if ModuleIO datatype matches record 'type' oneof name

    Args:
        io: a IO definition
        record: a record

    Returns:
        True if the right field is set, false otherwise

    See Also:
        :func:`fix_io_fmt` -- helper to translate message specification to oneof field name
    """
    return bool(record) and fix_io_fmt(io.message_format) == ubii.proto.TopicDataRecord.pb(record).WhichOneof('type')


class Scheduler(util.CoroutineWrapper):
    """
    A Scheduler is a wrapper around the coroutine created by its :meth:`get_trigger_loop` method.
    A scheduler -- depending on the used :attr:`.mode` -- waits for a certain delay and / or until
    any or all of its inputs are available and then schedules its callback.
    """

    DELTA_TIME_CACHE_SIZE = 30
    """
    number of cached delta times
    """

    def __init__(
            self,
            callback: typing.Callable[[], None],
            inputs: typing.Iterable[AsyncGetter],
            mode: ubii.proto.ProcessingMode,
            *,
            schedule_perf_metric: perf_calc_callable = functools.partial(perf_calc, 'schedule_delta_times'),
            exec_perf_metric: perf_calc_callable = functools.partial(perf_calc, 'exec_delta_times')

    ):
        """
        The mode and inputs determine when the callback is executed.

        Args:
            callback: Schedule execution of this callable when conditions are right
            inputs: depending on the :class:`~ubii.proto.ProcessingMode` inputs are part of the condition --
                or simply retrieved -- for execution of the callback. For each callable passed in ``inputs`` the
                returned :class:`~typing.Awaitable` will either be in :attr:`.done` or :attr:`.pending` when
                the callback is scheduled
            mode: used :class:`~ubii.proto.ProcessingMode`
            schedule_perf_metric: callable to evaluate scheduling overhead / performance, schould take one argument, the scheduler instance
            exec_perf_metric: callable to evaluate execution callback overhead / performance

        See Also:
            :attr:`ubii.proto.ProcessingMode.mode` -- details on processing modes
        """
        self._nursery = codestare.async_utils.TaskNursery(name=f'Nursery for {self}', loop=asyncio.get_running_loop())

        self.schedule_perf_metric: typing.Callable[[], float] = functools.partial(schedule_perf_metric, self)
        """
        Call this method to calculate the performance of the scheduling timings
        """
        self.exec_perf_metric: typing.Callable[[], float] = functools.partial(exec_perf_metric, self)
        """
        Call this method to calculate the performance of the execution
        """
        self.inputs: typing.Iterable[AsyncGetter] = inputs
        """
        callables to create awaitables for possibly needed inputs
        """
        self.schedule_delta_times: typing.Deque = collections.deque(maxlen=self.DELTA_TIME_CACHE_SIZE)
        """
        keep track of times between callback schedules for performance evaluation
        """
        self.exec_delta_times: typing.Deque = collections.deque(maxlen=self.DELTA_TIME_CACHE_SIZE)
        """
        keep track of execution times for callback for performance evaluation
        """
        self.done: typing.List[typing.Awaitable] = []
        """
        contains awaitables created from ``inputs`` argument that finished when callback needs to be scheduled
        """
        self.mode: 'ubii.proto.ProcessingMode' = mode
        """
        used mode, determines which conditions need to be matched to schedule the callback
        """
        self.callback: typing.Callable[[], None] = callback
        """
        callback to be scheduled
        """
        self.executor: 'concurrent.futures.ThreadPoolExecutor' = concurrent.futures.ThreadPoolExecutor(
            max_workers=MAX_WORKERS
        )
        """
        callbacks are possibly non-async callables, will be scheduled in this executor using 
        :meth:`asyncio.loop.run_in_executor`
        """
        self.task_clean_frequency = 20
        """
        After every n callback schedules, cancel old input awaitables
        """

        self._timing_epsilons = None
        self._timing_adjustment = 0.

        self._stop_during_next_iteration = False
        self._old_input_tasks = []
        self._scheduling_count = 0
        self._scheduling_timestamp = None
        self._exec_timestamp = None
        self._get_timestamp = self._nursery.loop.time

        super().__init__(coroutine=self._trigger_loop())

    @property
    def timing_thresholds(self) -> typing.Tuple[float, ...]:
        """
        These are the thresholds and minimum times for the asyncio sleep call. They depend on your system and
        event loop implementation. If they are set, the scheduler will adjust it's scheduling behaviour in frequency
        mode accordingly. If they are not set, no adjustments are made.

        Example:

            On linux with the default event loop you can set

                scheduler.timing_epsilons = (0.001, 0.00015)

            to adjust for the behaviour of the `epoll_await` system call

        """
        return self._timing_epsilons

    @timing_thresholds.setter
    def timing_thresholds(self, value) -> None:
        self._timing_epsilons = tuple(value)
        self._timing_adjustment = self._calculate_adjustment(self.delay)

    def _calculate_adjustment(self, delay):
        if 'frequency' not in self.mode:
            return 0.

        values = sorted((delay,) + self._timing_epsilons)
        cutoff = values.index(delay)
        if values[:cutoff]:
            return values[:cutoff][-1]

        return self._timing_epsilons[0] if self._timing_epsilons else 0.

    @property
    def timing_adjustment(self):
        """
        This read only property caches the appropriate adjustment for the timing code according to the
        set :attr:`timing_thresholds` and the processing mode (in all modes but frequency mode, the adjustment
        is 0)
        """
        return self._timing_adjustment

    @property
    def delay(self) -> float:
        """
        Delay in seconds to schedule callback, depending on :attr:`.mode`

            *   `frequency` -- simply use :attr:`~ubii.proto.ProcessingMode.Frequency.hertz`
            *   `trigger_on_input` -- use :attr:`~ubii.proto.ProcessingMode.TriggerOnInput.min_delay_ms`

        Raises:
            NotImplementedError: when the mode is ``lockstep``
        """
        if self.mode.frequency:
            return 1.0 / self.mode.frequency.hertz
        if self.mode.trigger_on_input:
            return self.mode.trigger_on_input.min_delay_ms / 1000
        if self.mode.lockstep:
            raise NotImplementedError(f"Not implemented yet")

    def halt(self) -> None:
        """
        Call this method to stop the internal scheduler loop after the next iteration, which will finish the
        wrapped coroutine
        """
        self._stop_during_next_iteration = True

    def _wait(self):
        """
        Create input awaitables from :attr:`.inputs`, start possible coroutines as tasks

        Returns:
            awaitable that waits for inputs to become ready, depending on :attr:`.mode`
        """
        if self.mode.frequency:
            timeout = self.delay - (self._get_timestamp() - self._scheduling_timestamp)
            adjustment = self._calculate_adjustment(timeout)
            timeout = timeout - adjustment if timeout > adjustment else 0
        else:
            timeout = None

        awts = [
            self._nursery.create_task(awt) if asyncio.iscoroutine(awt) else awt
            for awt
            in [get() for get in self.inputs]
        ]

        return_when = (
            asyncio.ALL_COMPLETED
            if self.mode.trigger_on_input and self.mode.trigger_on_input.all_inputs_need_update else
            asyncio.FIRST_COMPLETED
        )
        return asyncio.wait(awts, timeout=timeout, return_when=return_when)

    async def _trigger_loop(self):
        """
        This method is used to generate the coroutine that is wrapped internally.
        There is no need to call this method manually, just `await` the scheduler.

        The created coroutine runs an internal loop that can be stopped using :meth:`.halt`, and schedules
        the :attr:`.callback` in the currently running `event loop` with the :attr:`.executor` whenever the
        :attr:`.inputs` are `ready` (this means that either `one` or `all` of the input awaitables finished,
        depending on the :attr:`.mode`) and / or the :attr:`.delay` has passed.
        """

        _ = self.timing_adjustment
        self._scheduling_timestamp = self._get_timestamp()

        with self.executor as pool:
            while not self._stop_during_next_iteration:

                self.done, self.pending = await self._wait()
                since_last_schedule = self._get_timestamp() - self._scheduling_timestamp
                remaining = self.delay - since_last_schedule
                adjustment = self._calculate_adjustment(remaining)

                if remaining > adjustment:
                    await asyncio.sleep(remaining - adjustment)
                    since_last_schedule = self._get_timestamp() - self._scheduling_timestamp

                self.schedule_delta_times.append(since_last_schedule)
                self._exec_timestamp = self._get_timestamp()
                self._scheduling_timestamp = self._get_timestamp()
                await self._nursery.loop.run_in_executor(pool, self.callback)
                self.exec_delta_times.append(self._get_timestamp() - self._exec_timestamp)

                self._scheduling_count += 1
                if self.task_clean_frequency == 1:
                    await self._clean_old_inputs()
                    continue

                self._old_input_tasks.extend(self.pending)
                if self._scheduling_count % self.task_clean_frequency == 0:
                    self._scheduling_count = 0
                    await self._clean_old_inputs()

            await self._clean_old_inputs()

    async def _clean_old_inputs(self):
        with contextlib.suppress(asyncio.CancelledError):
            for awaitable in self._old_input_tasks:
                awaitable.cancel()
                await awaitable


T = typing.TypeVar('T')


@util.dunder.repr('id', 'name', 'status')
class ProcessingRoutine(ubii.proto.ProcessingModule, metaclass=util.ProtoRegistry):
    """
    This adds a :class:`ProcessingProtocol` providing processing behaviour to a
    :class:`~ubii.proto.ProcessingModule` representation.
    Refer to the documentation of the :class:`ProcessingProtocol` for
    information related to processing behaviour.
    """

    __unique_key_attr__ = 'name'
    """
    See documentation of :class:`ubii.framework.util.ProtoRegistry` for more information
    """

    class rules:
        """
        Rules to validate the protobuf message
        """

        @staticmethod
        def validate_language(pm: ProcessingRoutine):
            """
            Only python modules can be run by the python client node.
            Sets language to :attr:`~ubii.proto.ProcessingModule.Language.PY` if language is not set.

            Args:
                pm: needs validation

            Raises:
                ValueError: if :attr:`~ubii.proto.ProcessingModule.language` is already set,
                    but not to :attr:`~ubii.proto.ProcessingModule.Language.PY`
            """
            if pm.language and pm.language != pm.Language.PY:
                raise ValueError(f"{pm} can only run Python processing modules. language {pm.language!r} specified.")

            pm.language = pm.Language.PY

        @staticmethod
        def validate_id(pm: ProcessingRoutine):
            """
            Check if :attr:`~ubii.proto.ProcessingModule.id` is present

            Args:
                pm: needs validation

            Raises:
                ValueError: if :attr:`~ubii.proto.ProcessingModule.id` is not set
            """
            if pm.id is None:
                raise ValueError(f"id of {pm} is not set")

    def __init__(self, mapping=None, eval_strings=False, **kwargs):
        """
        Args:
            mapping (Union[dict, Message]): A dictionary or message to be
                used to determine the values for this message.
            eval_strings (bool): If this flag is set or the framework is used in :func:`~ubii.framework.logging.debug`
                mode, fields of the protobuf message that have names ending with ``_stringified`` (e.g.
                :attr:`~ubii.proto.ProcessingModule.on_processing_stringified`) will be evaluated.
                Creating a :class:`~ProcessingRoutine` whose
                :attr:`~ubii.proto.ProcessingModule.language` is not :attr:`~ubii.proto.ProcessingModule.Language.PY`
                is not possible, see :func:`.validate`

            kwargs (dict): Keys and values corresponding to the fields of the
                message and other arguments passed to :class:`ubii.proto.ProcessingModule`
        """

        # we allow initialisation from ub.ProcessingModule Wrappers
        if isinstance(mapping, ubii.proto.ProcessingModule):
            mapping = ubii.proto.ProcessingModule.pb(mapping)

        self._protocol = ProcessingProtocol(pm=self)
        self._change_specs = self._protocol.state.condition

        super().__init__(mapping=mapping, **kwargs)

        # this might set the language, so make sure that it is run before the next check
        self.validate()

        # eval stringified
        if eval_strings or util.debug() and self.language == self.Language.PY:
            self._eval_string_funcs()

        self._local_output_topics = topics.TopicStore(
            functools.partial(topics.BasicTopic, task_nursery=self._protocol.task_nursery)
        )

        self._input_source_getter = None
        self._output_destination_getter = None

    @property
    def protocol(self):
        """
        Reference to the :class:`ProcessingProtocol` managing this routine
        """
        return self._protocol

    def _eval_string_funcs(self):
        suffix = '_stringified'
        string_fields = [
            name for name in ubii.proto.ProcessingModule.pb(self).DESCRIPTOR.fields_by_name
            if name.endswith(suffix) and getattr(self, name)
        ]
        definitions = map(functools.partial(getattr, self), string_fields)
        for name, source in zip(string_fields, definitions):
            created_name = name[:-len(suffix)]
            ns = {}
            exec(source, {}, ns)
            func = ns.get(created_name)
            if not func or not callable(func):
                raise ValueError(f"Evaluating {name} did not create a function named {created_name}")

            setattr(type(self), created_name, func)

    @property
    def local_output_topics(self) -> 'topics.TopicStore':
        """
        Container to interact with the output topics of the module (e.g. register callbacks on the topics).
        """
        return self._local_output_topics

    @property
    def change_specs(self) -> 'asyncio.Condition':
        """
        Coordinate access to the protobuf specs of this module. Whenever part of the wrapped protobuf message changes,
        this should be done while acquiring this condition.

        When the :attr:`~ubii.proto.ProcessingModule.status` or input / output mappings change as part of
        :meth:`running <.start>` this protocol, coroutines waiting for this condition are notified automatically.

        Example: ::

            def change_output_mapping(pm: ProcessingRoutine, io: ubii.proto.ModuleIO):
                async with pm.change_specs:
                    pm.outputs = [io]
                    pm.change_specs.notify_all()

        """
        return self._change_specs

    def input_getter(
            self, io: ubii.proto.ModuleIO
    ) -> AsyncGetter[ubii.proto.TopicDataRecord | ubii.proto.TopicDataRecordList] | None:
        if not self._input_source_getter:
            return None

        return self._input_source_getter(io)

    def output_setter(self, io: ubii.proto.ModuleIO) -> AsyncSetter | None:
        if not self._output_destination_getter:
            return None

        return self._output_destination_getter(io)

    async def apply_io_mapping(self,
                               io_mapping: ubii.proto.IOMapping,
                               remote_topic_map: typing.Mapping[str, topics.Topic]):
        """
        Extract relevant information from a :class:`ubii.proto.IOMapping` message to initialize the
        :attr:`.get_output_topic` and :attr:`.get_input_topic` callables.

        Notifies awaitables waiting on :attr:`.change_specs` unless the mapping was empty.

        Args:
            io_mapping: Mapping that should be applied to this module
            remote_topic_map: While output topics will be managed by :attr:`.local_output_topics`, you need to provide
                a mapping for input topics to look up the topic source of the
                :attr:`ubii.proto.IOMapping.input_mappings`
        """
        get_input_mapping = {mapping.input_name: mapping for mapping in io_mapping.input_mappings}.get
        get_output_mapping = {mapping.output_name: mapping for mapping in io_mapping.output_mappings}.get

        def getter(topic_map: typing.Mapping[..., topics.Topic],
                   io: ubii.proto.ModuleIO) -> typing.Callable[[], typing.Awaitable]:
            """
            Depending on the :class:`~ubii.proto.InputMapping` for the IO specification,
            return a callable that returns an awaitable to get the next record in a topic,
            or the :attr:`~TopicMuxer.records` of a muxer

            Args:
                topic_map: This topic map will manage the topics that we access
                io: use this specification to look up the :class:`~ubii.proto.InputMapping` in the
                    :class:`~ubii.proto.IOMapping` passed to :meth:`apply_io_mapping`

            Returns:
                a callable returning an awaitable to await new topic data
            """
            mapping: ubii.proto.TopicInputMapping = get_input_mapping(io.internal_name)
            if mapping is None:
                raise ValueError(f"No mapping with input name {io.internal_name} found")

            if mapping.topic_mux:
                muxer = topics.TopicMuxer.registry.get(mapping.topic_mux.id)
                if not muxer:
                    muxer = topics.TopicMuxer(mapping=mapping.topic_mux)
                    topic_map[muxer.topic_selector].register_callback(lambda record: muxer.records.set([record]))

                return functools.partial(muxer.records.get, wait_for_write=not self.processing_mode.frequency)

            elif mapping.topic:
                return functools.partial(
                    topic_map[mapping.topic].buffer.get,
                    predicate=functools.partial(check_datatype, io),
                    wait_for_write=not self.processing_mode.frequency
                )
            else:
                raise ValueError(f"Invalid input mapping {mapping}")

        def setter(topic_map: typing.Mapping[..., topics.Topic],
                   io: ubii.proto.ModuleIO) -> typing.Callable[..., typing.Awaitable[None]]:

            """
            Depending on the :class:`~ubii.proto.OutputMapping` for the IO specification,
            return a callable that returns an awaitable to set the record in a topic (using a :class:`TopicDemuxer`
            is necessary)

            Args:
                topic_map: This topic map will manage the topics that we access
                io: use this specification to look up the :class:`~ubii.proto.OutputMapping` in the
                    :class:`~ubii.proto.IOMapping` passed to :meth:`apply_io_mapping`

            Returns:
                a callable returning an awaitable to set topic data
            """
            mapping: ubii.proto.TopicOutputMapping = get_output_mapping(io.internal_name)
            if mapping is None:
                raise ValueError(f"No mapping with input name {io.internal_name} found")

            async def set_value(record: ubii.proto.TopicDataRecord, default_topic=mapping.topic) -> None:
                """
                Set data for topic from the topic map, if the records datatype matches the datatype
                specified in the :class:`~ubii.proto.ModuleIO` specification

                Args:
                    record: a topic data record that should be written to a topic
                    default_topic: the default topic that should be used if the record has no topic set

                """
                assert isinstance(record, ubii.proto.TopicDataRecord)

                if not fix_io_fmt(io.message_format) in record:
                    log.warning(f"{record} does not have the {fix_io_fmt(io.message_format)} field set required"
                                f" by the ModuleIO data format, skipping")
                    return

                record.topic = record.topic or default_topic
                topic = topic_map[record.topic]

                # if the previous call creates a new topic with default callbacks, setting the record instantly
                # might not trigger the callback, so we wait until the topic buffer is ready
                # then write the data
                await topic.buffer.set(record, wait_for_read=True)

            if mapping.topic_demux:
                demuxer = topics.TopicDemuxer.registry.get(mapping.topic_demux.id)
                if not demuxer:
                    demuxer = topics.TopicDemuxer(mapping=mapping.topic_demux)

                async def set_value_demux(records: typing.List) -> None:
                    """
                    When demuxing, the :class:`TopicDemuxer` converts the record objects
                    to actual :class:`ubii.proto.TopicDataRecord` messages

                    Args:
                        records: objects that represent a record with additional meta information
                    """
                    for record in demuxer.convert_record_objects(records).elements:
                        await set_value(record)

                return set_value_demux
            elif mapping.topic:
                return set_value
            else:
                raise ValueError(f"Invalid output mapping {mapping}")

        # input is retrieved from remote topic map, output topics are created locally for each processing module
        self._input_source_getter = functools.partial(getter, remote_topic_map)
        self._output_destination_getter = functools.partial(setter, self.local_output_topics)

        if io_mapping.output_mappings or io_mapping.input_mappings:
            async with self.change_specs:
                self.change_specs.notify_all()

    @util.hook
    @util.document_decorator(util.hook)
    def on_created(self, context: types.SimpleNamespace) -> None:
        """
        Will be executed by the :class:`ProcessingProtocol` whenever
        :meth:`ProcessingProtocol.on_created` is called

        Args:
            context: Same context that is used by :meth:`ProcessingProtocol.on_created`

        """

    @util.hook
    @util.document_decorator(util.hook)
    def on_init(self, context: types.SimpleNamespace) -> None:
        """
        Will be executed by the :class:`ProcessingProtocol` whenever
        :meth:`ProcessingProtocol.on_init` is called

        Args:
            context: Same context that is used by :meth:`ProcessingProtocol.on_init`

        """

    @util.hook
    @util.document_decorator(util.hook)
    def on_processing(self, context: types.SimpleNamespace) -> None:
        """
        Will be executed by the :class:`ProcessingProtocol` whenever
        :meth:`ProcessingProtocol.on_processing` is called

        Args:
            context: Same context that is used by :meth:`ProcessingProtocol.on_processing`

        """

    @util.hook
    @util.document_decorator(util.hook)
    def on_halted(self, context: types.SimpleNamespace) -> None:
        """
        Will be executed by the :class:`ProcessingProtocol` whenever
        :meth:`ProcessingProtocol.on_halted` is called

        Args:
            context: Same context that is used by :meth:`ProcessingProtocol.on_halted`

        """

    @util.hook
    @util.document_decorator(util.hook)
    def on_destroyed(self, context: types.SimpleNamespace) -> None:
        """
        Will be executed by the :class:`ProcessingProtocol` whenever
        :meth:`ProcessingProtocol.on_destroyed` is called

        Args:
            context: Same context that is used by :meth:`ProcessingProtocol.on_destroyed`

        """

    def validate(self):
        """
        Run all rules in :attr:`.validation_rules`
        """

        for rule in self.validation_rules:
            rule(self)

    @classmethod
    @util.hook
    @util.document_decorator(util.hook)
    async def mutate_pm(cls, module: ProcessingRoutine, specs: ubii.proto.ProcessingModule) -> None:
        """
        Change specifications of a processing routine, but notify waiters for the modules
        :attr:`.change_specs` condition.

        Args:
            module: a processing routine
            specs: specifications as a :class:`ubii.proto.ProcessingModule` message
        """
        async with module.change_specs:
            # MergeFrom appends outputs and inputs, CopyFrom overwrites stuff, so we first need to "merge"
            # with custom logic (overwriting all non-falsy values)
            changes = {
                attr: value for attr, value in ubii.proto.ProcessingModule.to_dict(specs).items() if
                attr in specs
            }

            # then overwrite specs
            ubii.proto.ProcessingModule.copy_from(module, changes)
            module.change_specs.notify_all()

    @classmethod
    @util.hook
    @util.document_decorator(util.hook)
    async def start(cls, pm: ProcessingRoutine) -> ProcessingRoutine:
        """
        Start the internal :class:`ProcessingProtocol` of the passed routine.

        Args:
            pm: instance that needs to be started

        Returns:
            routine passed as ``pm``
        """
        assert pm.name in cls.registry
        pm._protocol.start()
        return pm

    @classmethod
    @util.hook
    @util.document_decorator(util.hook)
    async def stop(cls, pm: ProcessingRoutine):
        """
        Stop the internal :class:`ProcessingProtocol` of the passed routine and
        remove the routine from the registry.

        Args:
            pm: instance that needs to be stopped

        Returns:
            routine passed as argument
        """
        assert cls.registry.get(pm.name) == pm
        await pm._protocol.stop()
        async with pm.change_specs:
            await pm.change_specs.wait_for(
                lambda: pm.status == ubii.proto.ProcessingModule.Status.DESTROYED
            )
        assert pm._protocol.finished
        return pm

    @classmethod
    @util.hook
    @util.document_decorator(util.hook)
    async def halt(cls, pm: ProcessingRoutine):
        """
        Halt the internal :class:`ProcessingProtocol` of the passed routine.
        If th protocol is already in end state, do nothing but raise a warning.

        Args:
            pm: instance that needs to be halted

        Returns:
            routine passed as argument
        """
        if pm._protocol.finished:
            warnings.warn(f"Can't halt {pm}, protocol already finished")
            return pm

        assert pm.name in cls.registry
        await pm._protocol.state.set(PM_STAT.HALTED)

        async with pm.change_specs:
            await pm.change_specs.wait_for(
                lambda: pm.status == ubii.proto.ProcessingModule.Status.HALTED)
        return pm

    def __str__(self):
        info = f"name={self.name!r}"
        return f"{self.__class__.__name__}({info})"

    validation_rules: 'typing.List[typing.Callable[[ProcessingRoutine], None]]' = [
        rules.validate_language,
        rules.validate_id,
    ]
    """
    Callables to validate the protobuf message used in :meth:`.validate`
    
    See Also:
        :class:`ProcessingRoutine.rules` -- details on rules used here
    """

    def __delattr__(self, item):
        if item in type(self).to_dict(self):
            super().__delattr__(item)
        else:
            object.__delattr__(self, item)


class PM_STAT(enum.IntFlag):
    """
    Proxy for :class:`~ubii.proto.ProcessingModule.Status` but as :class:`enum.IntFlag`, to allow
    defining combinations of states.
    """
    INITIALIZED = enum.auto()
    CREATED = enum.auto()
    PROCESSING = enum.auto()
    HALTED = enum.auto()
    DESTROYED = enum.auto()


class ProcessingProtocol(protocol.AbstractProtocol[PM_STAT]):
    """
    This :class:`~ubii.framework.protocol.AbstractProtocol` implementation defines the Protocol used to run
    :class:`ProcessingRoutines <ProcessingRoutine>`.

    It defines valid :attr:`.state_changes` and callbacks,
    as well as the :attr:`.starting_state` and :attr:`.end_state`

    The :class:`.pm_proxy` methods are used to decorate the lifecycle callbacks
    :meth:`.on_created`, :meth:`.on_init`, :meth:`.on_processing`, :meth:`.on_halted`, :meth:`.on_destroyed`,
    to execute the proxied methods in the :class:`~ProcessingRoutine` which owns
    the :class:`ProcessingProtocol` and set the
    :class:`ProcessingRoutine.status`

    Attributes:
        task_nursery: inherited from :class:`ubii.framework.protocol.AbstractProtocol`
        state: inherited from :class:`ubii.framework.protocol.AbstractProtocol`
    """

    starting_state = PM_STAT.INITIALIZED
    """
    Before the protocol has the starting state, it's :attr:`.state` is `None`
    """
    end_state = PM_STAT.DESTROYED
    """
    If the protocol ends up in its end state, the coroutine `running` the protocol will finish
    """
    AnyState = PM_STAT.INITIALIZED | PM_STAT.CREATED | PM_STAT.PROCESSING | PM_STAT.HALTED | PM_STAT.DESTROYED
    """
    This is a combination of all possible states to allow easier transitions in :attr:`.state_changes`
    """

    class pm_proxy:
        """
        Define some decorators for callables like :meth:`~ProcessingProtocol.on_processing`, because
        all lifecycle callbacks need to set the :attr:`~ProcessingRoutine.status` of the corresponding
        processing routine, and call the corresponding callbacks (e.g. if a :meth:`ProcessingProtocol.on_processing` callback
        is called, it needs to call the :meth:`ProcessingRoutine.on_processing` callback of the
        :attr:`managed processing routine <ProcessingProtocol.pm>`)
        """
        _Status = typing.Union[PM_STAT, ubii.proto.ProcessingModule.Status]

        @classmethod
        def _get_decorator(cls, callable_: typing.Callable) -> typing.Callable:
            def decorator(func):
                @functools.wraps(func)
                async def _inner(instance: ProcessingProtocol, context):
                    result = callable_(instance.pm, context)
                    if asyncio.iscoroutine(result):
                        await result
                    return await func(instance, context)

                return _inner

            return decorator

        @classmethod
        def set_status_in_pm(cls, status: ubii.proto.ProcessingModule.Status) -> typing.Callable:
            """
            Callable decorated with returned decorator sets the status of the
            :class:`~ProcessingRoutine` instance to the passed status when called,
            and notifies awaitables waiting for :attr:`~ProcessingRoutine.change_specs`

            Args:
                status: which status will be set

            Returns:
                Decorator
            """

            async def set_status(pm: ProcessingRoutine, _):
                async with pm.change_specs:
                    pm.status = status
                    pm.change_specs.notify_all()

            return cls._get_decorator(set_status)

        @classmethod
        def callback_in_pm(cls, name: str) -> typing.Callable:
            """
            Callable decorated with returned decorator calls the method with specified name of the
            :class:`~ProcessingRoutine` instance.
            arguments of decorated callable are passed on.

            Args:
                name: e.g. ``'on_destroyed'``

            Returns:
                Decorator
            """

            def callback(pm: ProcessingRoutine, *args):
                _cb = getattr(pm, name)
                assert callable(_cb)
                result = _cb(*args)
                return result

            return cls._get_decorator(callback)

    class helpers:
        """
        Used for pre- / post-processing for the context data passed between lifecycle callbacks
        """

        __no_output__ = object()

        @classmethod
        def write_scheduler_data_to_context(cls, scheduler: Scheduler, ctx: types.SimpleNamespace) -> None:
            """
            Before processing, the processing context needs to be
            enriched with the topic data that the :class:`Scheduler` retrieved as inputs

            Args:
                scheduler: The Scheduler scheduling the callback
                ctx: The context that will be passed to the callback
            """
            ctx.delta_time = scheduler.schedule_delta_times[-1] if scheduler.schedule_delta_times else None
            inputs = vars(ctx.inputs)

            inputs.update(**{
                result.meta: result.value
                for result in map(lambda task: task.result(), scheduler.done)
            })
            ctx.inputs = types.SimpleNamespace(**inputs)

        @classmethod
        def publish_outputs_to_topics(cls, pm: ProcessingRoutine, ctx: types.SimpleNamespace):
            """
            Looks up topics via :meth:`~ProcessingRoutine.get_output_topic`,
            starts a async task to set the buffer value of the topic to the computed value from the
            context outputs, which triggers publishing if the topics has been set up correctly,
            see e.g. :meth:`ubii.node.node_protocol.implement_processing`.

            Args:
                pm: look up output topics in this routines
                    :meth:`~ProcessingRoutine.local_output_topics`
                ctx: extract computed outputs from this context

            """
            get_output_io = {io.internal_name: io for io in pm.outputs}.get

            for name, value in vars(ctx.outputs).items():
                if value == ProcessingProtocol.helpers.__no_output__:
                    continue

                io = get_output_io(name)
                if not io:
                    log.warning(f"No output topic for argument {name!r} defined in {pm}. Defined output handler[s] "
                                f"{', '.join(map(lambda o: repr(o.internal_name), pm.outputs))}.")
                    continue

                ctx.nursery.create_task(
                    pm.output_setter(io)(value)
                )

    def __init__(self, pm: ProcessingRoutine):
        """
        The created :class:`ProcessingProtocol` implements the behaviour of a :class:`~ubii.proto.ProcessingModule`
        or rather a :class:`ProcessingRoutine` during its lifetime. Each :class:`ProcessingProtocol` belongs to
        one :class:`ProcessingRoutine`

        Args:
            pm: the :class:`ubii.proto.ProcessingModule` wrapper that owns this protocol
        """
        super().__init__()
        self.pm: ProcessingRoutine = pm
        """
        reference to the :class:`~ubii.proto.ProcessingModule` wrapper that owns this protocol 
        """
        self.created_tasks: typing.List[asyncio.Task] = []
        """
        All tasks created by this protocol, to be stopped / cancelled when necessary
        """

    @pm_proxy.callback_in_pm('on_init')
    @pm_proxy.set_status_in_pm(ubii.proto.ProcessingModule.Status.INITIALIZED)
    async def on_init(self, context: types.SimpleNamespace):
        """
        First lifecycle method called after :attr:`pm` has been initialized

        Calls :meth:`pm.on_init() <ProcessingRoutine.on_init>` and sets the
        :attr:`~ProcessingRoutine.status` to :attr:`ub.ProcessingModule.Status.INITIALIZED`.

        Creates the following attributes in the context:

        -   ``context.loop`` the asyncio loop running the processing
        -   ``context.nursery`` the entity responsible for starting async tasks (either the ``context.loop`` or a
            reference to :attr:`.task_nursery`
        -   ``context.trigger_processing`` asyncio Event used later, see :meth:`.on_created`
        -   ``context.scheduler`` a :class:`Scheduler` instance, responsible for
            scheduling processing calls. Scheduled callback when inputs are available will set
            ``context.trigger_processing``
        -   ``context.muxer`` a reference to the :class:`TopicMuxer` type, needed to extract information
            from the records

        Finally, sets the :attr:`~ProcessingProtocol.status` to
        :attr:`PM_STAT.CREATED`

        Args:
            context: namespace object to hold data
        """

        # use the Processing Protocol as a task nursery if possible ?
        context.loop = asyncio.get_running_loop()
        context.nursery = self.task_nursery or context.loop
        context.muxer = topics.TopicMuxer
        context.demuxer = topics.TopicDemuxer

        # callable to create input_mapping dict from module io iterable
        make_input_dict: util.make_dict[typing.Tuple[str, str], AsyncGetter] = util.make_dict(
            key=lambda io: io.internal_name,
            value=self.pm.input_getter,
            filter_none=True
        )

        # wait for applied input mapping
        async with self.pm.change_specs:
            inputs = await self.pm.change_specs.wait_for(functools.partial(make_input_dict, self.pm.inputs))

        invalid_inputs = [input_values for input_values in inputs if not all(input_values)]
        if invalid_inputs:
            fmt = "{{ name: {!r}, format: {!r} }}"
            raise ValueError(f"Got invalid input[s]: {', '.join(map(lambda inp: fmt.format(*inp), invalid_inputs))}")

        # wait for applied output publishing
        async with self.pm.change_specs:
            waiter = self.pm.change_specs.wait_for(
                lambda: all(topic.callback_tasks for topic in self.pm.local_output_topics.values())
            )
            await asyncio.wait_for(waiter, timeout=10 if debug() else None)

        # start scheduler task
        context.trigger_processing = asyncio.Event()
        context.scheduler = Scheduler(
            callback=context.trigger_processing.set,
            inputs=[util.enrich(name, getter) for name, getter in inputs.items()],
            mode=self.pm.processing_mode
        )

        if 'linux' in sys.platform and 'frequency' in context.scheduler.mode:
            context.scheduler.timing_thresholds = (0.001, 0.00015)

        self.created_tasks += [context.nursery.create_task(context.scheduler)]
        await self.state.set(PM_STAT.CREATED)

    @pm_proxy.callback_in_pm('on_created')
    @pm_proxy.set_status_in_pm(ubii.proto.ProcessingModule.Status.CREATED)
    async def on_created(self, context):
        """
        Calls :meth:`pm.on_created() <ProcessingRoutine.on_created>` and sets the
        :attr:`~ProcessingRoutine.status` to :attr:`ub.ProcessingModule.Status.CREATED`.

        Prepares the context for further processing:
        -   creates inputs (`context.inputs`)
        -   Prepare output mapping (`context.outputs`)

        Waits for ``context.processing_trigger`` then extracts inputs from ``context.scheduler`` (using
        :meth:`.helpers.write_scheduler_data_to_context`) and changes
        the :attr:`~ProcessingProtocol.status` to
        :attr:`PM_STAT.PROCESSING`
        """
        log.info(f"created processing module {self.pm}")
        # create inputs
        context.inputs = types.SimpleNamespace(**dict.fromkeys(map(lambda io: io.internal_name, self.pm.inputs)))

        # create outputs (use dataclass to get better error reporting)
        fields = [
            (out.internal_name, typing.Any, dataclasses.field(default=self.helpers.__no_output__))
            for out in self.pm.outputs
        ]
        outputs = dataclasses.make_dataclass('outputs', fields, init=True)  # noqa
        context.outputs = outputs()

        # wait until processing is triggered and change state to processing
        await context.trigger_processing.wait()
        self.helpers.write_scheduler_data_to_context(context.scheduler, context)
        await self.state.set(PM_STAT.PROCESSING)

    @pm_proxy.callback_in_pm('on_processing')
    @pm_proxy.set_status_in_pm(ubii.proto.ProcessingModule.Status.PROCESSING)
    async def on_processing(self, context):
        """
        Calls :meth:`pm.on_processing() <ProcessingRoutine.on_processing>` and sets the
        :attr:`~ProcessingRoutine.status` to
        :attr:`ubii.proto.ProcessingModule.Status.PROCESSING`.

        -   publishes computed outputs (see :meth:`.helpers.publish_outputs_to_topics`)
        -   adjusts the ``context.scheduler`` callback (previously used to simply trigger processing) to

            -   extract data from context using :meth:`.helpers.write_scheduler_data_to_context`
            -   compute with :meth:`~ProcessingRoutine.on_processing` of owner
            -   publish outputs to topics using :meth:`.helpers.publish_outputs_to_topics`

            such that as long as the :attr:`~ProcessingProtocol.status` does not
            change data will be processed.

        Does not change the :attr:`~ProcessingProtocol.status` of the protocol,
        state change has to be triggered externally in this state.
        """

        self.helpers.publish_outputs_to_topics(self.pm, context)

        # we just change the scheduler callback to get inputs and publish outputs.
        context.scheduler.callback = util.function_chain(
            functools.partial(self.helpers.write_scheduler_data_to_context, context.scheduler, context),
            functools.partial(self.pm.on_processing, context),
            functools.partial(self.helpers.publish_outputs_to_topics, self.pm, context),
        )

    @pm_proxy.callback_in_pm('on_halted')
    @pm_proxy.set_status_in_pm(ubii.proto.ProcessingModule.Status.HALTED)
    async def on_halted(self, context):
        """
        Calls :meth:`pm.on_halted() <ProcessingRoutine.on_halted>` and sets the
        :attr:`~ProcessingRoutine.status` to :attr:`ub.ProcessingModule.Status.HALTED`.

        :meth:`Halts <Scheduler.halt>` the ``context.scheduler``

        Does not change the :attr:`~ProcessingProtocol.status` of the protocol,
        state change has to be triggered externally in this state.
        """
        context.scheduler.halt()

    @pm_proxy.callback_in_pm('on_destroyed')
    @pm_proxy.set_status_in_pm(ubii.proto.ProcessingModule.Status.DESTROYED)
    async def on_destroyed(self, context):
        """
        Calls :meth:`pm.on_destroyed() <ProcessingRoutine.on_destroyed>` and sets the
        :attr:`~ProcessingRoutine.status` to :attr:`ub.ProcessingModule.Status.DESTROYED`.

        Awaits cancellation of  all tasks that have been started by this protocol.
        """
        for task in self.created_tasks:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

    state_changes = {
        (None, PM_STAT.INITIALIZED): on_init,
        (PM_STAT.INITIALIZED, PM_STAT.CREATED): on_created,
        (PM_STAT.CREATED, PM_STAT.PROCESSING): on_processing,
        (AnyState, PM_STAT.HALTED): on_halted,
        (AnyState, PM_STAT.DESTROYED): on_destroyed,
    }
    """
    Possible state changes and respective callbacks
    """



from __future__ import annotations

import abc
import copy
import functools
import logging
import pprint
from typing import (
    TypeVar,
    Generic,
    Iterator,
    Mapping,
    MutableMapping,
    Union,
)

import ubii.proto
from . import (
    util,
    debug,
)
from .util.typing import Protocol

__protobuf__ = ubii.proto.__protobuf__
log = logging.getLogger(__name__)


class ServiceConnection(abc.ABC):
    """
    A ServiceConnection is a two-way request-reply connection. If the connection is present
    the only defined public API method is supposed to send a message in the ServiceRequest format,
    await a response from the master node and return it to the caller.
    """

    @abc.abstractmethod
    async def send(self, request: ubii.proto.ServiceRequest) -> ubii.proto.ServiceReply:
        """
        Send a `ServiceRequest` message, and receive a `ServiceReply`, according to the protobuf specifications.
        Can be implemented with any communication transports the master node supports.

        Args:
            request: ServiceRequest protobuf message (wrapper)
        """
        ...


class ServiceCall(ubii.proto.Service, metaclass=ubii.proto.ProtoMeta):
    """
    A ServiceCall is a callable that can be represented as a :class:`ubii.proto.Service` protobuf message.
    """

    def __init__(self, mapping=None, *, transport: ServiceConnection, **kwargs):
        """
        Args:
            mapping: for :class:`ubii.proto.Service` initialization
            transport: connection used to transfer messages
            **kwargs: for :class:`ubii.proto.Service` initialization
        """
        # we allow initialisation from ub.Service wrappers
        if isinstance(mapping, ubii.proto.Service):
            mapping = ubii.proto.Service.pb(mapping)

        super().__init__(mapping=mapping, **kwargs)
        self._transport = transport

    @property
    def transport(self):
        """
        Reference to the connection used
        """
        return self._transport

    @util.hook
    @util.document_decorator(util.hook)
    async def __call__(self, **payload) -> ubii.proto.ServiceReply:
        """
        send the :class:`ubii.proto.ServiceRequest` defined by the ``payload`` keyword arguments and the
        :attr:`.topic` of this :class:`ServiceCall` using the :attr:`.transport`.

        The reply is awaited and error handling is applied.

        Example:

            You can get :class:`ServiceCall` objects from a running client's :class:`~ubii.framework.client.Services`
            behaviour ::

                >>> from ubii.node import connect_client
                >>> from ubii.framework.client import Services
                >>> import asyncio
                >>> service_call = None
                >>> reply = None
                >>> async def main():
                ...     global service_call
                ...     global reply
                ...     async with connect_client() as client:
                ...             assert client.implements(Services)
                ...             service_call = client[Services].service_map.server_config
                ...             reply = await service_call()
                ...
                >>> asyncio.run(main())
                >>> service_call
                topic: "/services/server_configuration"
                response_message_format: "ubii.servers.Server"
                >>> reply
                server {
                  id: "f491e47e-e591-4961-b4ca-8e1142175ae8"
                  name: "master-node"
                  ...
                }

        Args:
            **payload: will be converted to a :class:`ubii.proto.ServiceRequest` message with
                :attr:`~ubii.proto.ServiceRequest.topic` defined by this calls :attr:`.topic`, i.e.
                typically only one keyword argument will be used, to define the field for the
                :attr:`~ubii.proto.ServiceRequest.type` oneof group. Values can be mappings or protobuf wrappers.

        Returns:
            the reply from the `master node`

        Raises:
            ubii.framework.errors.UbiiError: if `master node` replies with an error message

        :meta public:
        """
        request = ubii.proto.ServiceRequest(topic=self.topic, **payload)
        log.debug(f"sending service request:\n{request}")
        reply = await self._transport.send(request)
        log.debug(f"received service reply:\n{reply}")
        if reply.error:
            raise reply.error
        else:
            return reply

    @classmethod
    def register_decorator(cls, decorator, instance: object | None = None) -> None:
        """
        Since :meth:`.__call__` is a :class:`~ubii.util.hook`, you can easily decorate it.
        There are two ways to do that -- either directly use the fact that :meth:`.__call__` is a
        :class:`~ubii.util.hook` ::

            from ubii.framework.services import ServiceCall
            ServiceCall.__call__.register_decorator(decorator)

        Or use this method for convenience ::

            from ubii.framework.services import ServiceCall
            ServiceCall.register_decorator(decorator)

        Args:
            decorator: Some callable to decorate :meth:`.__call__`
            instance: only register decorator for this instance -- *optional*

        See Also:
            :class:`ubii.framework.util.functools.hook` -- more info about the hook decorator
        """
        cls.__call__.register_decorator(decorator, instance=instance)


T_Service = TypeVar('T_Service', bound=ServiceCall)
T_Service_Cov = TypeVar('T_Service_Cov', bound=ServiceCall, covariant=True)


class ServiceCallFactory(Protocol[T_Service_Cov]):
    def __call__(self: ServiceCallFactory[T_Service_Cov], mapping: ubii.proto.Service) -> T_Service_Cov:
        """
        :class:`ServiceCallFactory` objects need to have this call signature

        Args:
            mapping: instance of a :class:`ubii.proto.Service` wrapper for a `Service` message

        Returns:
            A type of :class:`ServiceCall` object
        """


class ServiceMap(ubii.proto.ServiceList, Mapping[str, T_Service], Generic[T_Service],
                 metaclass=ubii.proto.ProtoMeta):
    """
    A :class:`ServiceMap` is a wrapper around a `ServiceList` proto message  which
    provides a mapping for the :class:`~ubii.proto.Service` messages by topic.

    An adapter like this is needed because the master node advertises its services in a
    :class:`ubii.proto.ServiceList` message, but the semantics are to access them by topic
    -- typically with the `default topics` provided by the master node.

    Additionally, a :class:`ServiceMap` converts the respective :class:`ubii.proto.Service` for a topic to a
    :class:`ServiceCall` (the `service_call_factory` argument can be supplied during creation, it has to be a callable
    with signature matching a :class:`ServiceCallFactory`, and will be called with the :class:`~ubii.proto.Service`
    message as only argument). The results are cached.

    If you change the protobuf contents of a :class:`ServiceMap` object after creation (e.g. when you get a new
    :class:`~ubii.proto.ServiceList` from the master node), make sure to invalidate the cache by calling
    :meth:`.cache_clear`. This is automatically done if you assign to the :attr:`.elements` attribute directly,
    but if you e.g. copy the contents of a :class:`~ubii.proto.ServiceList` message into a
    :class:`ServiceMap` by means of :meth:`~proto.message.Message.copy_from`, you need to invalidate the cache manually.

    Until the message format changes and the `master node` advertises its services in a mapping, use this adapter.

    Example:

        .. _service_map_example:

        Create a :class:`ServiceMap` -- the example sets the `transport` for the :class:`ServiceCall` to `None`,
        real code needs to use a valid :class:`ServiceConnection` ::

            >>> from ubii.framework.services import ServiceMap, ServiceCall
            >>> from functools import partial
            >>> from ubii.proto import Service
            >>> services = [Service(topic=topic) for topic in ['foo', 'bar', 'foobar']]
            >>> service_map = ServiceMap(service_call_factory=partial(ServiceCall, transport=None), elements=services)

        When accessing keys that have corresponding services in the :attr:`.elements` of the map,
        a :class:`ServiceCall` is returned ::

            >>> ...
            >>> service = service_map['foo']
            >>> type(service)
            <class 'ubii.framework.services.ServiceCall'>
            >>> service.topic
            'foo'

        Services that are not defined in the :attr:`.elements` raise a :class:`KeyError` as expected ::

            >>> ...
            >>> service_map['thing']
            Traceback <...>
            KeyError: 'found no services for topic thing'

        The created :class:`ServiceCall` will be cached ::

            >>> ...
            >>> service is service_map['foo']
            True

        If you assign a `Service` list to :attr:`.elements` the cache will be cleared implictily and a new
        but equivalent :class:`ServiceCall` is created for the topic ::

            >>> ...
            >>> service_map.elements = services
            >>> service is service_map['foo']
            False
            >>> service == service_map['foo']
            True

        The :attr:`.elements` can change implicitly if the internal protobuf message changes. In this case
        the code also needs to call :meth:`.cache_clear` explicitly ::

            >>> ...
            >>> service_map.elements = services
            >>> from ubii.proto import ServiceList
            >>> service_list = ServiceList(elements=[service for service in services if not service.topic == 'foo'])
            >>> type(service_list).copy_from(service_map, service_list)
            >>> service is service_map['foo']
            True
            >>> service_map.elements
            [
            topic: "bar",
            topic: "foobar"
            ]

        As you can see the `foo` service is cached, although no `foo` topic is in the :attr:`.elements` anymore.
        After cache invalidation, accessing ``'foo'`` topic raises a :class:`KeyError` as expected ::

            >>> ...
            >>> service_map.elements = services
            >>> service_map.cache_clear()
            >>> service is service_map['foo']
            Traceback <...>
            KeyError: 'found no services for topic foo'
    """

    def __deepcopy__(self, memo):
        """
        Not provided by the proto plus base class. Needed because proto plus base
        implements __getattr__, so when you try to deepcopy a `ServiceMap` (e.g. :meth:`dataclasses.dataclass.replace`
        with a `dataclass` that contains a :class:`ServiceMap`) you get infinite recursion if you don't also implement
        `__deepcopy__`
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    def __init__(self: ServiceMap[T_Service],
                 mapping=None,
                 *,
                 service_call_factory: ServiceCallFactory[T_Service],
                 **kwargs):
        """

        Args:
            mapping: used to initialize the protobuf wrapper, can also itself be a :class:`~ubii.proto.ServiceList`
                wrapper
            service_call_factory: used to convert the protobuf messages to :class:`ServiceCall` objects
            **kwargs: passed to protobuf wrapper initialization
        """
        # we allow initialisation from ub.ServiceList Wrappers
        if isinstance(mapping, ubii.proto.ServiceList):
            mapping = ubii.proto.ServiceList.pb(mapping)

        super().__init__(mapping=mapping, **kwargs)

        # apply the lru cache wrapper per instance since the ServiceMap itself is not hashable
        self._make_service_call = functools.lru_cache(maxsize=128, typed=False)(
            functools.partial(self._get_service_call, service_call_factory)
        )

    def _get_service_call(self: 'ServiceMap[T_Service]', default_factory, topic) -> T_Service:
        found = [service for service in self.elements if service.topic == topic]
        if not len(found) == 1:
            raise KeyError(f"found {found or 'no services'} for topic {topic}")

        return default_factory(found[0])

    def cache_clear(self):
        """
        Clear cached service calls. Implicitly used if a new list of :class:`Services <~ubii.proto.Service>` is assigned
        to the :attr:`.elements` field, need to call manually if assignment is done implicitly.

        See Also:
            :ref:`ServiceMap code example<service_map_example>` -- shows caching behaviour in detail
        """
        self._make_service_call.cache_clear()

    def __setattr__(self, key, value):
        if key == 'elements':
            self.cache_clear()

        super().__setattr__(key, value)

    def __getitem__(self: 'ServiceMap[T_Service]', topic: Union[ubii.proto.ProtoField, str]) -> T_Service:
        return self._make_service_call(topic)

    def __len__(self) -> int:
        return len(self.elements)

    def __iter__(self) -> Iterator[str]:
        return (service.data for service in self.elements)


class DefaultServiceMap(ServiceMap[T_Service]):
    """
    Automatically creates Services for missing topics (like a :obj:`~collections.defaultdict`)
    Takes and optional mapping :math:`name \\rightarrow topic`, which can be accessed
    as attributes of the ServiceMap.

    Example:
        Make a map with defaults ::

            >>> from ubii.framework.services import ServiceCall, DefaultServiceMap
            >>> from functools import partial
            >>> service_map = DefaultServiceMap(
            ...     service_call_factory=lambda service: ServiceCall(topic=service.topic, transport=None),
            ...     defaults = {'foo': 'services/foo', 'bar': 'services/bar'},
            ... )

        Now the services for which there are defined ``defaults`` can be accessed as an attribute, and will
        be automatically added to the :attr:`.elements` ::

            >>> ...
            >>> service_map.elements
            []
            >>> service = service_map.foo
            >>> type(service)
            <class 'ubii.framework.services.ServiceCall'>
            >>> service.topic
            'services/foo'
            >>> service_map.elements
            [topic: "services/foo"]

        If no default is set, an :class:`AttributeError` is raised as expected ::

            >>> ...
            >>> service_map.boo
            Traceback <...>
            AttributeError: Unknown field for DefaultServiceMap: boo

        If the `framework` is used in debug mode, the :class:`AttributeError` contains information about possible
        matches in the :attr:`.defaults` ::

            >>> ...
            >>> from ubii.framework import debug
            >>> debug(True)
            True
            >>> service_map.boo
            Traceback <...>
            AttributeError: Unknown field for DefaultServiceMap: boo

            The above exception was the direct cause of the following exception:
            Traceback <...>
            AttributeError: DefaultServiceMap has no attribute 'boo'. Best match[es] in default topics: 'foo'

        services are only automatically created when the key is present as a value in :attr:`.defaults` ::

            >>> ...
            >>> service_map['bar']
            Traceback <...>
            KeyError: 'found no services for topic boo'
            >>> assert service_map['services/bar']
            True

    """

    def __init__(self, *args, defaults: MutableMapping[str, str] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._defaults = defaults or {}

    @property
    def defaults(self) -> MutableMapping[str, str]:
        """
        Reference to the mapping that was passed as ``defaults`` during initialization.

        Basically, attribute access is equivalent to item access for the corresponding `value` inside this
        mapping (if present) ::

            service_map = DefaultServiceMap(...)
            assert 'some_attr' in service_map.defaults
            service_map.some_attr == service_map[service_map.defaults['some_attr']]

        """
        return self._defaults

    def _get_service_call(self: DefaultServiceMap[T_Service], default_factory, topic) -> T_Service:
        service_call = None

        try:
            service_call = super()._get_service_call(default_factory, topic)
        except KeyError as e:
            # no service found
            if topic not in self._defaults.values() and self._defaults:
                raise e

            # create service
            self.elements += [ubii.proto.Service(topic=topic)]
            self.cache_clear()
        finally:
            return service_call or super()._get_service_call(default_factory, topic)  # try again

    def __getattr__(self, item):
        if item in self._defaults:
            return self[self._defaults[item]]

        try:
            return super().__getattr__(item)
        except AttributeError as e:
            # we want to give some additonal info because maybe it's just a typo
            if debug():
                info = self._get_debug_info(missing_attr=item)
                raise AttributeError(info) from e
            else:
                raise

    def _get_debug_info(self, missing_attr):
        from ubii.framework.util import similar
        matches = similar(self._defaults, missing_attr, cutoff=0.5)

        info = f"{self.__class__.__name__} has no attribute {missing_attr!r}. "

        if matches:
            info += f"Best match[es] in default topics: {', '.join(map(repr, matches))}"

        return info

    def __repr__(self):
        kls = self.__class__
        elements = f"<{type(self.elements).__name__} message at {id(self.elements)}>"
        defaults = f"<{type(self.defaults).__name__} with {len(self.defaults)} items>"
        return f"<{kls.__module__}.{kls.__name__} object at {id(self)} [defaults={defaults}, elements={elements}]>"

    def __str__(self):
        return pprint.pformat({name: getattr(self, name) for name in self.defaults if getattr(self, name)})



from __future__ import annotations

from codestare.async_utils import (
    accessor,
    condition_property,
    make_async,
    CoroutineWrapper,
    TaskNursery,
    async_exit_on_exc,
    awaitable_predicate,
    RegistryMeta,
    Registry,
)

from .collections import (
    merge_dicts,
    MatchMappingMixin,
    DefaultHookMap
)
from .enum import EnumMatcher
from .functools import (
    similar,
    hook,
    registry,
    exc_handler_decorator,
    calc_delta,
    log_call,
    ProtoRegistry,
    function_chain,
    compose,
    make_dict,
    async_compose,
    enrich,
    AbstractAnnotations,
    document_decorator,
    dunder
)

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property  # noqa

__DEBUG__ = False


def debug(enabled: bool | None = None) -> bool:
    """
    Call without arguments to get current debug state, pass truthy value to set debug mode.

    Args:
        enabled: If passed, turns debug mode on or off

    Returns:
        debug value
    """
    global __DEBUG__
    if enabled is not None:
        __DEBUG__ = bool(enabled)

    return __DEBUG__


__all__ = (
    "accessor",
    "awaitable_predicate",
    "condition_property",
    "make_async",
    "CoroutineWrapper",
    "TaskNursery",
    "async_exit_on_exc",
    "RegistryMeta",
    "Registry",
    "similar",
    "hook",
    "registry",
    "exc_handler_decorator",
    "log_call",
    "ProtoRegistry",
    "function_chain",
    "compose",
    "make_dict",
    "merge_dicts",
    "async_compose",
    "enrich",
    "calc_delta",
    "AbstractAnnotations",
    "debug",
    "document_decorator",
    "dunder"
)



import asyncio
import functools
import uuid

import pytest
import ubii.proto

import ubii.framework.client
import ubii.framework.processing
import ubii.framework.topics
import ubii.node.pytest


@pytest.fixture
def topic_data(request):
    return [ubii.proto.TopicData(topic_data_record=value) for value in request.param]


FakeMuxer = functools.partial(ubii.framework.topics.TopicMuxer, id=str(uuid.uuid4()))


class TestMuxer:
    @pytest.mark.parametrize('topic_data', [
        [{'topic': 'test/foo', 'bool': n % 2 == 0} for n in range(10)]
    ], indirect=True)
    async def test_muxer_basics(self, topic_data):
        topic_muxer = FakeMuxer(data_type='bool')
        for data in topic_data:
            await topic_muxer.records.set([data.topic_data_record])

        assert len(topic_muxer.records.value) == 1
        assert topic_muxer.records.value[0] == topic_data[-1].topic_data_record

    @pytest.mark.parametrize('topic_data', [
        pytest.param([{'bool': n % 2 == 0} for n in range(10)], id='alternating_bools')
    ], indirect=True)
    @pytest.mark.parametrize(
        'topic, identity_match_pattern, identity',
        [
            ('test/foo-1234', r'test/(.*)', 'foo-1234',),
            ('test/foo/bar', r'.+', 'test/foo/bar'),
            ('test/foo-1234', r'\d+', '1234')
        ]
    )
    async def test_identity_match(self, topic_data, topic, identity_match_pattern, identity):
        topic_muxer = FakeMuxer(data_type='bool', identity_match_pattern=identity_match_pattern)
        for data in topic_data:
            data.topic_data_record.topic = topic
            await topic_muxer.records.set([data.topic_data_record])

        assert len(topic_muxer.records.value) == 1
        record = topic_muxer.records.value[0]
        assert record.metadata()['identity'] == identity


class TestMuxerProcessing:
    processing_module = ubii.proto.ProcessingModule(
        name="muxer-processing-module",
        processing_mode={'trigger_on_input': {'min_delay_ms': 0,
                                              'all_inputs_need_update': True}},

        inputs=[
            {
                'internal_name': 'muxer_inputs',
                'message_format': 'int32'
            },
        ],
        outputs=[
            {
                'internal_name': 'demuxer_outputs',
                'message_format': 'int32'
            }
        ],
        on_processing_stringified='\n'.join((
            'def on_processing(self, context):',
            '    import ubii.proto',
            '    outputs = []',
            '    for record in context.inputs.muxer_inputs:',
            '        metadata = record.metadata()',
            '        if record.int32 % 2 == 0:',
            '            outputs += [{"int32": record.int32, "output_params": (metadata["identity"],)}]',
            '        else:',
            '            outputs += [{"int32": record.int32, "output_params": ("dummy-1234",)}]',
            '    context.outputs.demuxer_outputs = outputs',
        )),
        language=ubii.proto.ProcessingModule.Language.PY,
    )

    session = ubii.proto.Session(
        name="Muxer Session",
        processing_modules=[processing_module],
        io_mappings=[
            {
                'processing_module_name': processing_module.name,
                'input_mappings': [
                    {
                        'topic_mux': {
                            'data_type': 'int32',
                            'topic_selector': '/muxer/*',
                            'identity_match_pattern': '(?:/muxer/([0-9a-z-]+))'
                        },
                        'input_name': processing_module.inputs[0].internal_name,
                    },
                ],
                'output_mappings': [
                    {
                        'topic_demux': {
                            'data_type': 'int32',
                            'output_topic_format': '/demuxer/{{#0}}',
                        },
                        'output_name': processing_module.outputs[0].internal_name,
                    }
                ]
            },
        ]
    )

    client = ubii.proto.Client(
        name='Muxer Client',
        processing_modules=[processing_module],
        is_dedicated_processing_node=True
    )

    session_spec = [(session,)]
    module_spec = [(processing_module,)]
    client_spec = [(client,)]

    @pytest.mark.parametrize('topic_data', [
        [{'int32': n} for n in range(10)]
    ], indirect=True)
    async def test_muxer_processing(self, client, session_spec, module_spec, topic_data):
        await client
        await client.implements(ubii.framework.client.Sessions)
        await client[ubii.framework.client.Sessions].start_session(session_spec)

        topic, = await client[ubii.framework.client.Subscriptions].subscribe_regex('/demuxer/*')
        received = []
        topic.register_callback(received.append)

        for data in topic_data:
            data.topic_data_record.topic = f"/muxer/{client.id}"
            await client[ubii.framework.client.Publish].publish(data.topic_data_record)
            await asyncio.sleep(0.02)  # stagger inputs

        assert len(received) == 5
        assert all(record.topic == f"/demuxer/{client.id}" for record in received)
        assert all(value in [record.int32 for record in received] for value in [0, 2, 4, 6, 8])
        await client[ubii.framework.client.Sessions].stop_session(session_spec)



from __future__ import annotations

import asyncio
import difflib
import functools
import inspect
import logging
import pickle
import re
import types
import typing
import warnings
import weakref
from collections import namedtuple

import sys

import codestare.async_utils
import ubii.proto as ub
from . import RegistryMeta
from .typing import S, T, ExcInfo


# helper function, equivalent to functools._unwrap_partial
def _unwrap_partial(func):
    while isinstance(func, functools.partial):
        func = func.func
    return func


def similar(choices: typing.Sequence, item: typing.Any, cutoff=0.70):
    """
    Use this e.g. if you think your users can't type <:

    Example:

        >>> >>> from ubii.framework import util
        >>> choices = ["Foo", "Bar", "Foobar", "Thing"]
        >>> util.similar(choices, "Thong")
        ['Thing']
        >>> util.similar(choices, "foo")
        []
        >>> util.similar(choices, "foo", cutoff=0.5)
        ['Foo']
        >>> util.similar(choices, "foo", cutoff=0.4)
        ['Foo', 'Foobar']


    Args:
        choices: sequence of items
        item: something not in ``choices`` but possibly 'similar'
        cutoff: threshold for similarity measure

    Returns:
        elements from ``choices`` with :math:`similarity(choice, item) > cutoff`

    See Also:
        :class:`difflib.SequenceMatcher` -- used internally to compute similarity score
    """
    _similarity = lambda k: difflib.SequenceMatcher(None, k, item).ratio()  # noqa
    return list(sorted(filter(lambda item: _similarity(item) > cutoff, choices), key=_similarity, reverse=True))


T_Callable = typing.TypeVar('T_Callable', bound=typing.Callable[..., typing.Any])


class append_doc:
    """
    Helper to append information to docstrings of callables
    """
    _whitespace = re.compile(r'^( +)[^\s]+', re.MULTILINE)

    def __init__(self, cb):
        self.cb = cb

    def __call__(self, info: str):
        doc = self.cb.__doc__ or ''
        orig_indents = self._whitespace.findall(doc) or ['']
        info_indents = self._whitespace.findall(info) or ['']

        # check if indents are consistent:
        assert all(
            indent.startswith(orig_indents[0]) for indent in orig_indents
        ), f"Inconsistent indents for {self.cb}.__doc__"
        assert all(
            indent.startswith(info_indents[0]) for indent in info_indents
        ), f"Inconsistent indents for {info}"

        replace = functools.partial(re.compile(f"^{info_indents[0]}", re.MULTILINE).sub, orig_indents[0])
        self.cb.__doc__ = doc + '\n'.join(map(replace, info.split('\n')))


def document_decorator(decorator):
    def inner(decorated):
        # if not os.environ.get('SPHINX_DOC_BUILDING'):
        #     return decorated
        is_async = asyncio.iscoroutinefunction(decorated)
        sig = inspect.signature(decorated)

        if is_async:
            ret_type = '[' + sig.return_annotation.strip("'") + ']' if sig.return_annotation != sig.empty else ''
            decorated.__annotations__['return'] = f"Awaitable{ret_type}"

        info = {
            'decorator':
                f"{decorator.__module__}.{decorator.__qualname__}" if not isinstance(decorator, str) else decorator,
            'signature':
                "{}def {}{}".format("async " if is_async else '', decorated.__name__, sig)
        }

        append_doc(decorated)(
            """
            This callable had the :obj:`~{decorator}` decorator applied.
            Original signature: ``{signature}``
            """.format(**info)
        )
        return decorated

    return inner


class calc_delta:
    """
    Calculate difference between values produced by a factory function

    Example:

        >>> from ubii.framework import util
        >>> value_factory = iter([1, 2, 70, 3, 4, 5]).__next__
        >>> delta = util.calc_delta(value_factory)
        >>> delta()
        1
        >>> delta()
        68
        >>> delta()
        -67
        >>> delta.value
        3

    """

    def __init__(self, get_value):
        self.get_value = get_value
        self.value = get_value()

    def __call__(self):
        previous = self.value
        self.value = self.get_value()
        return self.value - previous


class hook(typing.Generic[T_Callable]):
    """
    This decorator gives the decorated callable the ability to
    :meth:`register decorators <.register_decorator>`, i.e. define a consistent API to apply
    decorators to the decorated callable.

    Notes:

        *   Decorators need not to be unique, registering the same decorator multiple times will apply it multiple times.
        *   Decorators are applied at the first call to the :class:`hook`, i.e. the first call could take longer.
        *   :attr:`hook.func` is the unmodified callable passed during initialization. If necessary one can clear the
            cached 'applied' version of the callable with :meth:`hook.cache_clear`

    Warning:
        Decorator order depends on order of registration. The last registered decorator is applied last.
        We have :math:`hook = d_n \\circ d_{n-1} \\circ ... \\circ d_1 \\circ d_0 \\circ f` where
        :math:`0` to :math:`n` are the indices of the decorator in :attr:`hook.decorators`


    Example:

        You have some callable which is predestined to be slightly altered later. Instead of monkey patching it later,
        you can preemptively define that callable to be 'alterable' by converting it into a `hook`.

        >>> from ubii.framework import util
        >>> value_factory = iter([1, 2, 70, 3, 4, 5]).__next__
        >>> hook = util.hook(value_factory)
        >>> hook()
        1
        >>> hook()
        2
        >>> hook()
        70
        >>> def decorator(func):
        ...     def inner():
        ...             return -1 * func()
        ...     return inner
        ...
        >>> hook.register_decorator(decorator)
        >>> hook()
        -3
        >>> hook()
        -4

    """
    __name__: str

    def __init__(self: hook[T_Callable], func: T_Callable, decorators=None):
        """
        Create a hook from ``func``

        Args:
            func: a callable which should be more easily decorate-able
            decorators: initial decorators -- optional
        """
        self.func = func
        """
        Reference to unmodified callable
        """
        self._global_decorators = list(decorators or ())
        self._instance_decorators: typing.MutableMapping[int, typing.List] = {}
        self._instance_callables: typing.MutableMapping[int, typing.Callable] = {}
        self._applied = None
        functools.wraps(func)(self)

    def decorators(self, instance: object | None = None):
        """
        List of registered decorators

        Args:
            instance: return decorators for this specific instance -- `optional`
        """
        return self._global_decorators + (self._instance_decorators.get(id(instance), []) if instance else [])

    def register_decorator(self, decorator, instance: object | None = None) -> None:
        """
        Add decorator to internal list of decorators and re-apply all decorators at next call

        Args:
            decorator: some callable
            instance: only register decorator for this specific instance -- `optional`
        """
        if instance is not None:
            weakref.finalize(instance, lambda key: self._instance_decorators.pop(key, None), id(instance))
            self._instance_decorators.setdefault(id(instance), []).append(decorator)
            self._instance_callables.pop(id(instance), None)
        else:
            self._global_decorators.append(decorator)
            self.cache_clear()

    def cache_clear(self):
        """
        Apply decorators again at next access
        """
        self._applied = None

    def __call__(self, *args, __first_argument_is_instance=False, **kwargs):
        if self._applied is None:
            self._applied = compose(*self._global_decorators)(self.func) if self._global_decorators else self.func

        if args and __first_argument_is_instance:
            instance = args[0]
            if id(instance) not in self._instance_callables and id(instance) in self._instance_decorators:
                self._instance_callables[id(instance)] = compose(*self._instance_decorators[id(instance)])(self)
            else:
                self._instance_callables[id(instance)] = self

            func = self._instance_callables[id(instance)]
        else:
            func = self._applied

        return func(*args, **kwargs)

    def __get__(self, instance=None, owner=None):
        if instance is None:
            return self

        # we need to mangle the name
        special_args = {f"_{type(self).__name__}__first_argument_is_instance": True}

        return functools.partial(self, instance, **special_args)


class registry(typing.Generic[S, T]):
    """
    Decorator to register every call to another callable

    Example:

        Let's say we have another decorator that does some simple task

            >>> def my_decorator(func):
            ...     def inner(*args):
            ...         print("Foo")
            ...         return func(*args)
            ...     return inner

        We want to create a new decorator that does the same but keeps track of every
        decorated function

            >>> from ubii.framework import util
            >>> my_decorator_registry = util.registry(key=lambda func: func.__name__, fn=my_decorator)
            >>> @my_decorator_registry
            ... def test_function(foo: str):
            ...     print(foo)
            ...
            >>> my_decorator_registry.registry
            {'test_function': <function test_function at (...)>}
            >>> test_function("Bar")
            Foo
            Bar

    """

    def __init__(self: registry[S, T], key: typing.Callable[[T], S], fn: typing.Callable[..., T]):
        """
        This callable wraps the callable passed as `fn`, but caches results in :attr:`.registry`
        with `key(result)` as key.

        Args:
            key: computes some hashable unique value for possible results of wrapped callable
            fn: if this callable returns non-unique results, old cached values are technically overwritten since the
                ``key`` returns the same value for the same input by definition
        """
        self.key = key
        """
        computes keys for results of :attr:`.fn` to cache them inside :attr:`.registry`
        """
        self.fn = fn
        """
        wrapped callable
        """
        self._registry: typing.Dict[S, T] = {}
        functools.wraps(fn)(self)

    @property
    def registry(self) -> typing.Dict[S, T]:
        """
        Mapping :math:`key \\rightarrow result` from computed :attr:`key(result) <.key>` of all calls
        """
        return self._registry

    def __call__(self, *args, **kwargs):
        instance = self.fn(*args, **kwargs)
        self._registry[self.key(instance)] = instance
        return instance

    def __get__(self, instance=None, owner=None):
        if instance is None:
            return self

        return functools.partial(self, instance)


def exc_handler_decorator(handler: typing.Callable[[ExcInfo], typing.Awaitable[None | bool] | None | bool]):
    """
    This callable takes an 'exception handler' i.e. a callable that processes results of :func:`sys.exc_info`
    and converts it to a decorator that can be itself applied to ``async`` callables to handle their exceptions.

    Example:

        >>> def handler(*exc_info):
        ...     if exc_info:
        ...             print(f"Got exception {exc_info[1]}")

        >>> from ubii.framework import util
        >>> handler_deco = util.exc_handler_decorator(handler)
        >>> @handler_deco
        ... async def foo(value):
        ...     print(5 // value)

        >>> async def main():
        ...     for value in [0,1,2,0,-2]:
        ...             await foo(value)

        >>> import asyncio
        >>> asyncio.run(main())
        Got exception integer division or modulo by zero
        5
        2
        Got exception integer division or modulo by zero
        -3

    Args:
        handler: an exception handler callable

    Returns:
        decorator to catch exceptions from ``async`` methods

    """

    def decorator(fun):
        @functools.wraps(fun)
        async def _inner(*args, **kwargs):
            try:
                result = fun(*args, **kwargs)
                if asyncio.isfuture(result) or asyncio.iscoroutine(result):
                    result = await result
                return result
            except:  # noqa
                exception_info = sys.exc_info()
                result = handler(*exception_info)
                if asyncio.isfuture(result) or asyncio.iscoroutine(result):
                    await result

        return _inner

    return decorator


def log_call(logger: logging.Logger | None = None):
    """
    Args:
        logger: calls are logged as :meth:`logging.Logger.debug`, if not passed use logger with module name

    Returns:
        A decorator to log calls to decorated callable to ``logger``
    """

    def decorator(fun):
        log = logger or logging.getLogger(fun.__module__)

        @functools.wraps(fun)
        def __inner(*args):
            log.debug(f"called {fun}")
            return fun(*args)

        return __inner

    return decorator


class AbstractAnnotations:
    attr_name = '__required_annotations__'
    wrap_marker = object()

    def __init__(self, *names):
        self.names = names

    def __call__(self, cls: type):
        if any(name not in cls.__annotations__ for name in self.names):
            raise ValueError(f"{cls} needs to have annotation[s] for {', '.join(self.names)}")

        names = getattr(cls, self.attr_name, set())
        names.update(self.names)
        setattr(cls, self.attr_name, names)

        original_new = cls.__new__

        wrapped = self.wrap_once(original_new)  # might return original_new if already decorated
        setattr(cls, original_new.__name__, wrapped)

        return cls

    def wrap_once(self, original_new):
        wrap_markers = getattr(original_new, 'markers', set())
        if self.wrap_marker in wrap_markers:
            return original_new

        @functools.wraps(original_new)
        def wrapped(cls, *args, **kwargs):
            instance = original_new(cls, *args, **kwargs)

            # is instance can be created type(instance) can't have abstract methods, so
            # instance also has to have all required attributes
            missing = [name for name in getattr(cls, self.attr_name, []) if not hasattr(cls, name)]
            if missing:
                raise TypeError(
                    f"Can't create {cls} instance with missing class attribute[s] {', '.join(map(repr, missing))}")

            return instance

        wrap_markers.add(self.wrap_marker)
        wrapped.markers = wrap_markers
        return wrapped


class ProtoRegistry(ub.ProtoMeta, RegistryMeta):
    """
    Instances for types that have this metaclass are registered, and can be serialized / deserialized
    according to their protobuf specifications.

    See Also:
        :class:`~codestare.async_utils.helpers.RegistryMeta` -- the :class:`ProtoRegistry` simply adds the serialization
            to the mechanisms for registration of instances inherited from here
    """

    def __new__(mcs, *args, **kwargs):
        kls = super().__new__(mcs, *args, **kwargs)
        return kls

    def _serialize_all(cls):
        return {key: cls.serialize(obj) for key, obj in cls.registry.items()}

    def _deserialize_all(cls, mapping: typing.Mapping):
        return {key: cls.deserialize(obj) for key, obj in mapping.items()}

    def save_specs(cls, path):
        """
        Serialize all registered Protocol Buffer Wrapper objects and pickle the results to ``path``
        """
        with open(path, 'wb') as file:
            pickle.dump(cls._serialize_all(), file)

    def update_specs(cls, path) -> None:
        """
        Loads specs from pickled file, updates all registered instances according to their specification
        from the pickled messages.

        Args:
            path: load binary pickle file from here

        """
        with open(path, 'rb') as file:
            loaded = pickle.load(file)

        specs = cls._deserialize_all(loaded)
        for key, item in cls.registry.items():
            spec = specs.get(key)
            if not spec:
                warnings.warn(f"No {cls} instance for key {key} registered, can't update")
                continue

            cls.copy_from(item, spec)


class function_chain:
    """
    Generates a callable that calls multiple functions in a defined order with same arguments

    Example:

        >>> def foo(value):
        ...     print(f"foo: {value}")
        >>> def bar(value):
        ...     print(f"bar: {value}")
        >>> from ubii.framework import util
        >>> chain = util.function_chain(foo, bar)
        >>> chain(1)
        foo: 1
        bar: 1

    See Also:
        :class:`compose` -- if you want to compose functions instead of passing the same arguments to each

        :class:`async_compose` -- if you want to compose coroutines
    """

    def __init__(self, *funcs):
        self.funcs = funcs
        """
        Tuple of functions that need to be called
        """

    def __call__(self, *args):
        for f in self.funcs:
            f(*args)

    def __get__(self, instance=None, owner=None):
        if instance is None:
            return self

        return functools.partial(self, instance)

    @classmethod
    def reverse(cls, *funcs):
        return cls(*reversed(funcs))


class compose:
    """
    Generates a callable that is the composition of other callables.
    Callables are called in the order in which they are passed to :class:`compose` i.e.
    :math:`compose(f, g) = g \\circ f`

    Example:
        >>> from ubii.framework import util
        >>> def foo(value):
        ...     return value + 1
        ...
        >>> def bar(value):
        ...     print(value)
        ...
        >>> foobar = util.compose(foo, bar)  # foobar(value) = bar(foo(value))
        >>> foobar(1)
        2

    See Also:
        :class:`function_chain` -- if you want to pass the same argument to every function instead of composing

        :class:`async_compose` -- if you want to compose coroutines

    """

    def __init__(self, *funcs):
        self._info = ', '.join(map(repr, funcs))
        if not funcs:
            raise ValueError(f"No callables passed to compose")

        self.funcs = funcs
        """
        Tuple of original callables
        """
        self._reduced = functools.reduce(lambda g, f: lambda *a: f(g(*a)), funcs)

    def __call__(self, *args):
        return self._reduced(*args)

    def __repr__(self):
        return f"compose({self._info})"


class make_dict(typing.Generic[S, T]):
    """
    This callable creates dictionaries using a key-function and a value-function
    to extract information from the values produced by some iterable.

    Example:

        >>> from ubii.framework import util
        >>> from collections import namedtuple
        >>> from random import sample
        >>> Foo = namedtuple('Foo', ['id', 'value'])
        >>> def random_foo(k):
        ...     return [Foo(*v) for v in zip(range(k), sample(range(k), k))]
        ...
        >>> id_to_value = util.make_dict(key=lambda f: f.id, value=lambda f: f.value)
        >>> some_foos = random_foo(5)
        >>> some_foos
        [Foo(id=0, value=2), Foo(id=1, value=0), Foo(id=2, value=1), Foo(id=3, value=3), Foo(id=4, value=4)]
        >>> id_to_value(some_foos)
        {0: 2, 1: 0, 2: 1, 3: 3, 4: 4}
        >>> more_foos = random_foo(3)
        >>> more_foos
        [Foo(id=0, value=0), Foo(id=1, value=2), Foo(id=2, value=1)]
        >>> id_to_value(more_foos)
        {0: 0, 1: 2, 2: 1}


    """

    def __init__(self: make_dict[S, T],
                 key: typing.Callable[[typing.Any], S],
                 value: typing.Callable[[typing.Any], T],
                 filter_none=False):
        self._key = key
        self._value = value
        self._filter = filter_none

    def __call__(self, iterable: typing.Iterable) -> typing.Dict[S, T]:
        return {
            self._key(item): self._value(item)
            for item in iterable
            if not self._filter or self._value(item)
        }


class async_compose:
    """
    Like :class:`compose` for coroutines.
    Coroutines are awaited in the order in which they are passed to :class:`async_compose` i.e.
    ``composed = async_compose(f, g)`` is equivalent to ``async def composed(*args): return await g(await f(*args))``

    Example:

        >>> from ubii.framework import util
        >>> import asyncio, random
        >>> async def simulate_IO(value):
        ...     await asyncio.sleep(random.random())
        ...     return value
        ...
        >>> async def make_values(k):
        ...     for n in range(k):
        ...             yield await simulate_IO(n)
        ...
        >>> async def process(values):
        ...     return [value async for value in values if value % 2 == 0]
        ...
        >>> async def result(values):
        ...     for value in values:
        ...             print(await simulate_IO(value))
        ...
        >>> processed = util.compose(make_values, process)  # make_values does not return a coroutine -> compose
        >>> composed = util.async_compose(processed, result) # processed returns a coroutine -> async_compose
        >>> asyncio.run(composed(5))
        0
        2
        4


    See Also:
        :class:`function_chain` -- if you want to pass the same argument to every function instead of composing

        :class:`compose` -- if you want to compose normal callables instead of coroutines

    """

    class composed:
        def __init__(self,
                     f: typing.Callable[..., typing.Awaitable],
                     g: typing.Callable[..., typing.Awaitable]):
            self.f = f
            self.g = g

        async def __call__(self, *args):
            return await(self.g(await self.f(*args)))

        def __repr__(self):
            return f"{self.g!r}({self.f!r})(...)"

    def __init__(self, *fns):
        self._reduced = functools.reduce(self.composed, fns)

    def __call__(self, *args):
        return self._reduced(*args)

    def __repr__(self):
        return repr(self._reduced)


class enrich(typing.Callable[..., 'enrich.result']):
    """
    Enriches the results of a callable with some meta information.

    Useful if you deal with callables which you can't or don't want to change, to avoid
    leaking dependencies -- e.g. the part of the codebase that "knows" the meta information is not
    the same that defines the callables, and it's undesirable to change that.

    Works with normal callables and also coroutine functions -- if ``async`` coroutine functions / methods
    are used, the resulting callable will `await` the result first, before it adds the meta information and
    returns it to the caller.

    Example:

        Consider a mapping of callables:

        >>> import functools
        >>> make_print = functools.partial(functools.partial, print)
        >>> call_mapping = {value: make_print(value) for value in ['foo', 'bar', 'foobar']}
        >>> call_mapping['foo']()
        foo

        Pretend that ``make_print`` is defined somewhere else in your code, maybe even in someone else's code.
        You know that it's a factory for callables that print one value, and the code where you define the
        call mapping also knows which value each callable in the mapping will print.

        You don't want to pass this meta information around in your code (let's say as
        additional arguments for your calls), just because the implementation of ``make_print`` does not behave
        the way you want -- i.e. it does not also return the value that gets printed.

        So instead, you could define your mapping like this:

        >>> from ubii.framework import util
        >>> call_mapping = {v: util.enrich(v, make_print(v)) for v in ['foo', 'bar', 'foobar']}
        >>> result = call_mapping['foo']()
        foo
        >>> result
        result(value=None, meta='foo')

        Now your code can simply deal with the "fixed" callables.

    """
    result = namedtuple('result', ['value', 'meta'])
    """
    Named tuple for access of original return value and meta information
    """

    def __init__(self, meta, func: typing.Callable):
        """
        Add meta information to results of callable

        Args:
            meta: Something that each call to this callable should return in the :attr:`~.result.meta` field
            func: Simple callable or coroutine function. Result will be available as :attr:`~.result.value` field
        """
        assert callable(func), f"{func} needs to be a callable"
        self._reduced: typing.Union[async_compose, compose]

        # func opjects can be kinda complex, so to choose the right kind of "compose" we need to find
        # out what actually is happening
        # sadly inspect.iscoroutinefunction does not handle bound methods the way we need

        base = _unwrap_partial(func) if isinstance(func, functools.partial) else func
        base = base.__func__ if inspect.ismethod(base) else base

        if asyncio.iscoroutinefunction(base):
            async def attach(value):
                return self.result(value=value, meta=meta)

            self._reduced = async_compose(func, attach)
        else:
            def attach(value):
                return self.result(value=value, meta=meta)

            self._reduced = compose(func, attach)

    def __call__(self, *args):
        return self._reduced(*args)


class dunder:
    """
    the following decorators can be used to create dunder methods on classes.

    Note:

        Why not use `attrs <https://www.attrs.org/>`_?
            *   because the ubii framework should be usable by people that don't have a lot of experience in python, so
                it uses the standard library whenever possible -> dataclasses instead of `attrs`
            *   dataclasses are not useful for complex classes, nonetheless even a very complex class might have
                very easy `dunder` methods
    """

    @classmethod
    def all(cls, *attrs):
        """
        A decorator that applies all decorators from :class:`dunder`

        Args:
            *attrs: the names of attributes that should be part of the dunder methods

        Returns:
            a class decorator
        """
        return compose(cls.repr(*attrs), cls.hash(*attrs))

    @classmethod
    def repr(cls, *attrs, patch_str_for_builtins: bool = True):
        """
        Creates a __repr__ method on the class, that is basically ::

            def __repr__(self):
                info = {attr: getattr(self, attr, None) for attr in attrs}
                return f"{self.__class__.__name__}({', '.join('{}={}'.format(k, v) for k,v in info.items())})"

        Args:
            *attrs: the names of attributes that should be part of the __repr__
            patch_str_for_builtins: if True, types that inherit directly from builtins will receive a __str__ method
                as well (so set this to false if you implement __str__ yourself) -- **optional**

        Returns:
            a class decorator
        """

        def __repr__(instance):
            info = {attr: getattr(instance, attr, None) for attr in attrs}
            return (f"<{instance.__class__.__module__}.{instance.__class__.__name__} "
                    f"[{', '.join('{}={!r}'.format(k, v) for k, v in info.items())}]>")

        def decorator(kls):
            kls.__repr__ = functools.wraps(kls.__repr__)(__repr__)

            # builtin types use a C implementation for __str__ and patching __repr__ after the type has been
            # created will not make the C __str__ method fall back to the new __repr__ like expected, so we
            # need to explicitly patch the __str__ method of types that inherit the __str__ from builtins
            if isinstance(kls.__str__, types.WrapperDescriptorType) and patch_str_for_builtins:
                kls.__str__ = kls.__repr__

            return kls

        return decorator

    @classmethod
    def hash(cls, *attrs):
        """
        Creates a __hash__ method on the class, that is basically ::

            def __hash__(self):
                return hash((self.__class__, ) + tuple(map(lambda attr: getattr(self, attr, None), sorted(attrs))))

        Args:
            *attrs: the names of attributes that should be part of the __hash__

        Returns:
            a class decorator
        """

        def __hash__(instance):
            return hash((instance.__class__,) + tuple(map(lambda attr: getattr(instance, attr, None), sorted(attrs))))

        def decorator(kls):
            kls.__hash__ = functools.wraps(kls.__hash__)(__hash__)
            return kls

        return decorator


__all__ = (
    'similar',
    'hook',
    'registry',
    'exc_handler_decorator',
    'log_call',
    'ProtoRegistry',
    'function_chain',
    'compose',
    'make_dict',
    'async_compose',
    'enrich',
    'calc_delta',
    'dunder',
    'document_decorator',
    'append_doc'
)



"""
Concrete implementation of :class:`~ubii.framework.services.ServiceConnection` and
:class:`~ubii.framework.topics.DataConnection` using :mod:`aiohttp`.
"""

from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import functools
import json
import logging
import socket
import typing
import typing as t
import urllib.parse
import warnings

import aiohttp
import ubii.proto

from ubii.framework import (
    services,
    topics,
    util,
    debug
)

local_ip = socket.gethostbyname(socket.gethostname())
"""
global attribute containing the local ip of the machine

:meta hide-value:
"""
log = logging.getLogger(__name__)


@util.dunder.repr('url')
class AIOHttpConnection:
    """
    Base class for connections using :mod:`aiohttp`
    """

    def __init__(self, url: str, host_ip: str = local_ip):
        """
        Create connection to specific url

        Args:
            url: url that you want to connect to, make sure that it uses a valid url scheme
            host_ip: your local ip, needed to create the right CORS headers
        """
        self._session_is_set = asyncio.Event()
        self._session = None
        self.url: str = url
        """
        field to access url
        """
        self.https: bool = urllib.parse.urlparse(url).scheme == 'https'
        """
        True if url scheme of :attr:`.url` is 'https' else False
        """
        self.host_ip: str = host_ip
        """
        Used for :attr:`.headers`
        """

    @property
    def headers(self) -> typing.Dict[str, str]:
        """
        dictionary containing key ``origin`` with value equal to the host's tcp port url, to be
        included in the requests for CORS authentication with the `master node`

        Example:

            >>> from ubii.node import connect_client
            >>> import asyncio
            >>> async def main():
            ...     async with connect_client() as client:
            ...             topic_connection = client.protocol._context.topic_connection
            ...             assert topic_connection is not None
            ...             print(topic_connection.host_ip)
            ...             print(topic_connection.headers)
            ...
            >>> asyncio.run(main())
            '10.0.0.1'
            {'origin': 'http://10.0.0.1:8080'}

        """
        localhost_prefix = '127.0.0.'  # actually check for 127.0.0/8 some time
        host = 'localhost' if self.host_ip.startswith(localhost_prefix) else self.host_ip
        return {'origin': f"http{'s' if self.https else ''}://{host}:8080"}

    @property
    def session(self) -> aiohttp.ClientSession:
        """
        Session used for requests, can have special debug handling or JSON formatter

        When setting this property a private event is set, that can be used to "queue" requests until
        a session is defined.

        When this property is deleted, the event is unset i.e. future requests can wait for a new session
        to be set.
        """
        return self._session

    @session.setter
    def session(self, value: aiohttp.ClientSession):
        if value is None:
            if self._session_is_set.is_set():
                raise ValueError("Can't unset by setting to None. Delete the attribute instead.")
            else:
                return

        if self._session_is_set.is_set():
            warnings.warn(f"session is already set (see documentation for more info).")
            return

        self._session = value
        self._session_is_set.set()

    @session.deleter
    def session(self):
        self._session = None
        self._session_is_set.clear()


class AIOHttpWebsocketConnection(AIOHttpConnection, topics.DataConnection):
    """
    A simple WebSocket connection that implements the :class:`~ubii.framework.topics.DataConnection` abstract base class.
    """

    @dataclasses.dataclass
    class Events:
        """
        public events that are set and unset during the lifetime of the connection
        """
        connected: asyncio.Event = dataclasses.field(default_factory=asyncio.Event)
        disconnected: asyncio.Event = dataclasses.field(default_factory=asyncio.Event)

    log_socket_in = logging.getLogger(f"{__name__}.in.socket")
    """
    Logger for incoming traffic
    """
    log_socket_out = logging.getLogger(f"{__name__}.out.socket")
    """
    Logger for outgoing traffic
    """

    def __anext__(self) -> t.Awaitable[ubii.proto.TopicData]:
        return self._stream.__anext__()  # type: ignore

    def __init__(self, url: str, host_ip: str = local_ip, max_message_size: int | None = 0):
        """
        Create websocket connection

        Args:
            url: Url to connect to, according to broker config, see :class:`.AIOHttpConnection` for more info
            host_ip: ip of host, see :class:`.AIOHttpConnection` for more info
            max_message_size: maximum allowed size for socket reads,
                set to `None` to use aiohttp default (4MB), out default is 0 (no limit)
        """
        super().__init__(url, host_ip)
        self._ws: aiohttp.ClientWebSocketResponse | None = None
        self._client_id: str | None = None
        self._stream = self._stream_coro()
        self._max_message_size = max_message_size
        self.events: AIOHttpWebsocketConnection.Events = self.Events()
        """
        Public events to wait for connection / disconnection in client code
        """

    @contextlib.asynccontextmanager
    async def connect(self, client_id: str):
        """
        Use this async context manager to establish a connection for a specific client, and disconnect it
        afterwards.

        Warning:

            A :class:`AIOHttpWebsocketConnection` can only be connected with one client id.
            If you try to connect a connected connection, a warning will be raised.

        Example:

            Setup:
                *   Master node running on localhost
                *   local ip address ``10.0.0.1``

            Note that the url of the connection that is created by default (with :class:`~ubii.node.connect_client`)
            contains the local ip of the machine not ``localhost`` because the default protocol
            adjusts the connections' url  -- once the `master node` communicates its ip address -- to match the ip
            of the master node. On the one hand this makes sure that the `master node` sends a valid
            :class:`~ubii.proto.Server` message, on the other hand this makes output / logs more readable because all
            connections share the same ip in their url (the
            :attr:`~ubii.node.protocol.LegacyProtocol.Context.service_connection` is also adjusted once the
            `master node` sends the :class:`~ubii.proto.Server` message)

            >>> from ubii.node import connect_client
            >>> import asyncio
            >>> async def main():
            ...     async with connect_client() as client:
            ...         topic_connection = client.protocol._context.topic_connection
            ...         assert topic_connection is not None
            ...         async with topic_connection.connect(client.id) as connected:
            ...             print(connected.url)
            ...
            >>> asyncio.run(main())
            [...] UserWarning: <ubii.framework.connections.AIOHttpWebsocketConnection with url=ws://10.0.0.1:8104> is already connected.
            ws://10.0.0.1:8104

            It's also possible to handle messages from the topic connection manually (but not advisable) ::

                from ubii.node import connect_client
                import asyncio

                async def foo():
                    async with connect_client() as client:
                        topic_connection = client.protocol.context.topic_connection
                        assert topic_connection is not None
                        async for topic_data in topic_connection:
                            # do something
                            ...

        Args:
            client_id: e.g. content of a :attr:`ubii.proto.Client.id` field.

        Returns:
            an async context manager
        """
        if self.events.connected.is_set():
            warnings.warn(f"{self} is already connected.")
            yield self
        else:
            self.client_id = client_id
            kwargs = {'max_msg_size': self._max_message_size} if self._max_message_size is not None else {}
            async with self.session.ws_connect(f"{self.url}/?clientID={self.client_id}", **kwargs) as ws:
                self.ws = ws
                yield self

            del self.ws

    async def _stream_coro(self) -> ubii.proto.TopicData:
        await self.events.connected.wait()
        assert self.ws is not None
        message: aiohttp.WSMessage
        async for message in self.ws:
            if message.type == aiohttp.WSMsgType.ERROR:
                self.log_socket_in.error(message.data)
                if debug():
                    raise message.data

            elif message.type == aiohttp.WSMsgType.TEXT:
                if message.data == "PING":
                    await self.ws.send_str('PONG')
                    self.log_socket_out.log(logging.DEBUG - 5, f"Sending 'PONG'")

                self.log_socket_in.log(logging.DEBUG - 5, f"Received {message.data}")
            elif message.type == aiohttp.WSMsgType.BINARY:
                data = ubii.proto.TopicData.deserialize(message.data)
                self.log_socket_in.debug(f"Received {data}")
                yield data
            else:
                self.log_socket_in.warning(f"Unknown message Type for message: {message}")

        log.info(f"Closing Websocket connection")
        del self.ws

    @property
    def ws(self) -> aiohttp.ClientWebSocketResponse | None:
        """
        If connection is open use this :class:`~aiohttp.ClientWebSocketResponse` object
        to read and send messages

        Setting this attribute clears the :attr:`~.Events.disconnected` event, and sets the
        :attr:`~.Events.connected` event in :attr:`.events`

        Deleting this attribute clears the :attr:`~.Events.connected` event, and sets the
        :attr:`~.Events.disconnected` event in :attr:`.events` (unless already disconnected)
        """
        return self._ws

    @ws.setter
    def ws(self, value: aiohttp.ClientWebSocketResponse):
        if value is None:
            raise ValueError("Can't unset by setting to None. Delete the attribute instead.")

        if self.events.connected.is_set():
            warnings.warn(f"ws is already set ({self._ws}). Delete the attribute first (see documentation)")
            return

        self._ws = value
        self.events.disconnected.clear()
        self.events.connected.set()
        log.info(f"Connected {self}")

    @ws.deleter
    def ws(self):
        if not self._ws:
            return

        self._ws = None
        self.events.connected.clear()
        self.events.disconnected.set()
        log.info(f"Disconnected {self}")

    @property
    def client_id(self):
        """
        If set, the connection belongs to a specific client. If not set, the connection is not usable.

        Setting this attribute is only allowed if it is not already set. Otherwise, a warning will be raised.
        Delete the attribute to clear it and allow setting it to a different id.
        """
        return self._client_id

    @client_id.setter
    def client_id(self, value):
        if self._client_id:
            warnings.warn(f"client_id is already set ({self._client_id}). Unset first (see documentation)")
            return

        self._client_id = value

    @client_id.deleter
    def client_id(self):
        self._client_id = None

    async def send(self, data: ubii.proto.TopicData, timeout=None):
        """
        Use this coroutine to send :class:`~ubii.proto.TopicData` messages to the `master node`.
        Will wait until :attr:`.events.connected` is set i.e. until the connection has a
        valid url with client id.

        Args:
            data: message to send
            timeout: used for waiting for the :attr:`.events.connected` event, as well as for
                actually sending of the message i.e. for the first message the actual time until an error is
                raised can be at most double the timeout value (in seconds). If not set, coroutine waits
                an unlimited time -- `optional`

        Raises:
            asyncio.TimeoutError: if ``timeout`` is set and connection or sending of message takes longer
        """
        await asyncio.wait_for(self.events.connected.wait(), timeout=timeout)
        assert self.ws is not None
        self.log_socket_out.debug(f"Sending {data}")
        await asyncio.wait_for(self.ws.send_bytes(ubii.proto.TopicData.serialize(data)), timeout=timeout)


class AIOHttpRestConnection(AIOHttpConnection, services.ServiceConnection):
    """
    Send Service Request Messages
    """

    async def send(self, request: ubii.proto.ServiceRequest, timeout=None) -> ubii.proto.ServiceReply:
        """
        Send a :class:`~ubii.proto.ServiceRequest` and wait for :class:`~ubii.proto.ServiceReply` from `master node`.

        Args:
            request: as proto plus message wrapper object for a request
            timeout: if not set, wait indefinitely. Else wait for ``timeout`` seconds for the connection to become
                usable (needs a :attr:`~AIOHttpRestConnection.session`), then send the message and wait for ``timeout`` seconds for reply,
                then raise error

        Returns:
            a wrapper object for the reply

        Raises:
            asyncio.TimeoutError: if ``timeout`` is set and connection is not usable in time or reply is not received in time

        See Also:

            :attr:`ubii.proto.ServiceRequest.type` -- the oneof group for possible request fields

            :attr:`ubii.proto.ServiceReply.type` -- the oneof group for possible reply fields
        """
        await asyncio.wait_for(self._session_is_set.wait(), timeout=timeout)
        async with self.session.post(self.url, headers=self.headers, json=request, timeout=timeout) as resp:
            json = await asyncio.wait_for(resp.text(), timeout=timeout)
            return ubii.proto.ServiceReply.from_json(json,
                                                     ignore_unknown_fields=True)  # master node bug requires ignore


def aiohttp_session():
    """
    We create a aiohttp session with our custom json encoder and some logging handlings
    in debug mode.

    If :func:`ubii.framework.util.debug` is used to turn on debug mode of the framework,
    this function will return a different session, with logging for requests (see :class:`aiohttp.TraceConfig`)
    and a timeout of `1` second compared to `300` seconds for the normal session.

    Both sessions use the :class:`ubii.proto.util.ProtoEncoder` to serialize json, and raise Exceptions when
    requests fail (see :attr:`aiohttp.ClientSession.raise_for_status`)

    >>> from ubii.node import connections
    >>> from ubii.framework import debug
    >>> session = connections.aiohttp_session()
    >>> debug(True)
    True
    >>> debug_session = connections.aiohttp_session()
    >>> session.timeout
    ClientTimeout(total=300, connect=None, sock_read=None, sock_connect=None)
    >>> debug_session.timeout
    ClientTimeout(total=1, connect=None, sock_read=None, sock_connect=None)
    >>> session.trace_configs
    []
    >>> debug_session.trace_configs
    [<aiohttp.tracing.TraceConfig object at 0x7f657a080c70>]
    >>> session.raise_for_status
    True
    >>> debug_session.raise_for_status
    True
    >>> session.json_serialize
    functools.partial(<function dumps at 0x7f657b3f4940>, cls=<class 'ubii.proto.util.ProtoEncoder'>)
    >>> debug_session.json_serialize
    functools.partial(<function dumps at 0x7f657b3f4940>, cls=<class 'ubii.proto.util.ProtoEncoder'>)
    """
    if util.debug():
        trace_config = aiohttp.TraceConfig()

        async def on_request_start(session, context, params):
            logging.getLogger('aiohttp.client').debug(f'Starting request <{params}>')

        trace_config.on_request_start.append(on_request_start)
        trace_configs = [trace_config]
        timeout = aiohttp.ClientTimeout(total=1)
    else:
        timeout = aiohttp.ClientTimeout(total=300)
        trace_configs = []

    from ubii.proto import ProtoEncoder
    return aiohttp.ClientSession(raise_for_status=True,
                                 json_serialize=functools.partial(json.dumps, cls=ProtoEncoder),
                                 trace_configs=trace_configs,
                                 timeout=timeout)



from __future__ import annotations

import asyncio
import enum
import logging
import random
import warnings
from dataclasses import dataclass
from typing import Callable

import pytest

from ubii.framework.client import UbiiClient
from ubii.framework.protocol import AbstractProtocol, RunProtocol
from ubii.node.protocol import LegacyProtocol, States as UbiiStates

pytestmark = pytest.mark.asyncio
log = logging.getLogger(__name__)


class MockProtocol(AbstractProtocol[UbiiStates]):
    starting_state = UbiiStates.STARTING
    end_state = UbiiStates.STOPPED

    async def _on_start(self, context):
        log.info(context)
        context.data = "test"
        await self.state.set(UbiiStates.CONNECTED)

    async def _on_connect(self, context):
        log.info(context)
        assert context.data == 'test'

        async def stop_at_some_point():
            await asyncio.sleep(random.randrange(5, 10))
            await self.state.set(UbiiStates.STOPPED)

        context.stop_task = asyncio.create_task(stop_at_some_point())

    async def _on_stopped(self, context):
        log.info(context)
        context.stop_task.cancel()
        await context.stop_task

    state_changes = {
        (None, UbiiStates.STARTING): _on_start,
        (UbiiStates.STARTING, UbiiStates.CONNECTED): _on_connect,
        (UbiiStates.CONNECTED, UbiiStates.STOPPED): _on_stopped,
    }


async def test_mock_protocol():
    protocol = MockProtocol()
    coro = RunProtocol(protocol)
    await asyncio.create_task(coro)


@pytest.fixture
async def protocol(request):
    protocol: AbstractProtocol = request.param()
    running = protocol.start()
    yield running
    if not running.finished:
        await running.stop()


@pytest.fixture(autouse=True)
def warning_filter():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore',
            category=RuntimeWarning,
            module='ubii.node.protocol',
            message="Not setting the protocol client can lead to unexpected behaviour."
        )
        yield


@pytest.mark.parametrize('protocol', [LegacyProtocol], indirect=True)
async def test_default_protocol(protocol: LegacyProtocol):
    await asyncio.wait_for(
        protocol.state.get(predicate=lambda state: state == UbiiStates.CONNECTED),
        timeout=6
    )
    client = protocol.client
    assert client.id


@pytest.mark.parametrize('protocol', [LegacyProtocol], indirect=True)
async def test_default_protocol_stop(protocol: LegacyProtocol):
    await asyncio.wait_for(protocol.state.get(predicate=lambda state: state == UbiiStates.CONNECTED),
                           timeout=5)
    client = protocol.client
    assert client.id
    await protocol.stop()
    assert protocol.state.value == protocol.end_state


class TestProtocol(AbstractProtocol):
    class TestStates(enum.IntFlag):
        START = enum.auto()
        RUNNING = enum.auto()
        END = enum.auto()
        ANY = START | RUNNING | END

    async def on_start(self, context):
        print(f"starting with context:\n-> {context}")
        await asyncio.sleep(2)  # simulating some setup IO ...
        await self.state.set(self.TestStates.RUNNING)

    async def on_run(self, context):
        print(f"running with context:\n-> {context}")

    async def on_stop(self, context):
        print(f"stopping with context:\n-> {context}")

    starting_state = TestStates.START
    end_state = TestStates.END

    state_changes = {
        (None, TestStates.START): on_start,
        (TestStates.START, TestStates.RUNNING): on_run,
        (TestStates.ANY, TestStates.END): on_stop,
    }


@dataclass(init=True, repr=True, eq=True)
class FooBehaviour:
    foo: Callable[[str], None] | None = None


@dataclass(init=True, repr=True, eq=True)
class BarBehaviour:
    bar: str | None = None


class FakeClientProtocol(TestProtocol):
    TestStates = TestProtocol.TestStates

    def __init__(self):
        super().__init__()
        self.client: UbiiClient | None = None

    async def on_run(self, context):
        assert self.client is not None
        self.client[FooBehaviour].foo = print

    state_changes = {**TestProtocol.state_changes, (TestStates.START, TestStates.RUNNING): on_run}


async def test_fake_protocol():
    protocol = FakeClientProtocol()

    client = UbiiClient(
        name='Test Client',
        required_behaviours=(FooBehaviour,),
        optional_behaviours=(BarBehaviour,),
        protocol=protocol
    )
    protocol.client = client

    async def use_behaviours():
        await client.implements(BarBehaviour, FooBehaviour)
        client[FooBehaviour].foo(f"Using behaviours -> {client[BarBehaviour].bar}")

    async def implement_bar():
        client[BarBehaviour].bar = 'Bar'

    async with client as started_client:
        assert not started_client.implements(BarBehaviour)
        await asyncio.gather(implement_bar(), use_behaviours())



import asyncio
import logging
import random
import typing as t

import pytest
import ubii.proto as ub

from ubii.framework.topics import DataConnection, TopicStore, BasicTopic, StreamSplitRoutine

pytestmark = pytest.mark.asyncio
log = logging.getLogger(__name__)


class MockConnection(DataConnection):
    """
    Mock data published to topics 'A' and 'B'
    """

    def __init__(self, max_items_send: int):
        self._max_items_send = max_items_send
        self._async_gen_coro = self._gen_data()

    async def send(self, data: ub.TopicData):
        raise NotImplementedError

    def _gen_record(self):
        for i in range(self._max_items_send):
            topic = 'A' if i % 2 == 0 else 'B'

            if i % 2 == 0:
                yield ub.TopicDataRecord(topic=topic, bool=i % 3 != 0)
            else:
                yield ub.TopicDataRecord(topic=topic, int32=i)

    async def _gen_data(self):
        for record in self._gen_record():
            yield ub.TopicData(topic_data_record=record)
            await asyncio.sleep(random.randrange(1, 3) * 0.1)

    def __anext__(self) -> t.Awaitable[ub.TopicData]:
        return self._async_gen_coro.__anext__()


class Data(t.NamedTuple):
    container: TopicStore
    coro: StreamSplitRoutine


@pytest.fixture
def make_connection(request):
    """
    Create a StreamSplitRoutine from a MockConnection and a container, return topics 'A' and 'B' as well as the routine
    """

    def generate(*, max_items_send: int):
        return MockConnection(max_items_send=max_items_send)

    yield generate


@pytest.fixture
def container():
    container = TopicStore(BasicTopic)
    yield container


@pytest.mark.parametrize('items', [20, 1, 99])
async def test_topics(items, make_connection, container):
    """
    Test if splitting to topics and callbacks work correctly
    """
    a, b = container['A'], container['B']
    a_received = []
    b_received = []

    a.register_callback(a_received.append)
    b.register_callback(b_received.append)

    await asyncio.create_task(
        StreamSplitRoutine(container=container, stream=make_connection(max_items_send=items))
    )

    assert len(a_received) == (items + 1) // 2
    assert len(b_received) == items - len(a_received)

    assert [record.bool for record in a_received] == list(map(lambda i: i % 3 != 0, range(len(a_received))))
    assert [record.int32 for record in b_received] == list(filter(lambda x: x % 2, range(items)))


@pytest.mark.parametrize('items', [15])
async def test_task_manager(make_connection, container, items):
    """
    Test if the task_manager (in charge of canceling the tasks) works properly
    """
    a, b = container['A'], container['B']
    a_received = []
    b_received = []

    # register tasks with task manager, tasks for topic 'a' should be automatically unregistered after async with block
    async with a.task_nursery:
        a.register_callback(a_received.append)
        b.register_callback(b_received.append)
        await asyncio.create_task(
            StreamSplitRoutine(container=container, stream=make_connection(max_items_send=items))
        )

    assert all(t.cancelled() for t in a.callback_tasks.values())  # should be cancelled
    assert not any(t.cancelled() for t in b.callback_tasks.values())  # should not be cancelled

    # check results to see if tasks ran correctly and were not cancelled prematurely
    assert [record.bool for record in a_received] == list(map(lambda i: i % 3 != 0, range(len(a_received))))
    assert [record.int32 for record in b_received] == list(filter(lambda x: x % 2, range(items)))

    # reset the topic to be reused for further testing
    a.callback_tasks.clear()
    a.task_nursery = type(a.task_nursery)(name=a.task_nursery.name, loop=a.task_nursery.loop)

    # register again ...
    async with a.task_nursery:
        a.register_callback(a_received.append)
        # ... but change task_manager
        # previously this was a feature of a topic, now we do it explicitly
        b.task_nursery.push_async_exit(a.task_nursery.pop_all())

    assert not any(t.cancelled() for t in b.callback_tasks.values())  # should not be cancelled
    assert not any(
        t.cancelled() for t in a.callback_tasks.values()
    )  # should not be cancelled since task_manager was changed before block was closed

    await b.task_nursery.aclose()  # should cancel all tasks (also for topic 'a')
    assert all(t.cancelled() for t in b.callback_tasks.values())
    assert all(t.cancelled() for t in a.callback_tasks.values())


@pytest.mark.parametrize('items', [15])
async def test_task_regex_topics(make_connection, container, items):
    all_regex = container['*']
    all_regex_received = []
    all_regex.register_callback(all_regex_received.append)

    await asyncio.create_task(
        StreamSplitRoutine(container=container, stream=make_connection(max_items_send=items))
    )

    assert len(all_regex_received) == items

    a_regex = container['A*']
    a_regex_received = []
    a_regex.register_callback(a_regex_received.append)

    await asyncio.create_task(
        StreamSplitRoutine(container=container, stream=make_connection(max_items_send=items))
    )

    assert len(all_regex_received) == 2 * items
    assert len(a_regex_received) == (items + 1) // 2



import asyncio
import contextlib
import functools
import os
import threading

import itertools
import pytest

import ubii.framework.processing

DEFAULT_TIMEOUT = 5


@pytest.fixture
def event_loop(request, record_property) -> asyncio.AbstractEventLoop:
    """
    Since the test in this file closes the event loop, we need
    to create a new one each time
    """
    record_property('event_loop_scope', request.scope)
    loop = asyncio.get_event_loop_policy().new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()


class FakePM(ubii.framework.processing.ProcessingRoutine):
    """
    Test PM
    """

    def __init__(self,
                 context,
                 argument=None,
                 **kwargs):
        super().__init__(**kwargs)
        self._argument = argument

    @property
    def argument(self):
        return self._argument


class TestCLI:
    @pytest.fixture
    def cancel_after_timeout(self, request):
        def cancel(timeout):
            import time
            import signal
            time.sleep(timeout)
            signal.raise_signal(signal.SIGINT)

        thread = threading.Thread(target=functools.partial(cancel, request.param))
        thread.daemon = True
        thread.start()
        yield

    @pytest.mark.parametrize(
        'args, cancel_after_timeout',
        [
            pytest.param(
                [f"--debug --processing-modules {FakePM.__module__}.{FakePM.__qualname__}"], DEFAULT_TIMEOUT,
                id="test_pm"
            ),
            pytest.param(
                [
                    "--debug",
                    f"--processing-modules test:{FakePM.__module__}.{FakePM.__qualname__}",
                    "--module-args test:argument='foo',name='test'"
                ], DEFAULT_TIMEOUT,
                id="test_mod_args"
            )
        ],
        indirect=['cancel_after_timeout']
    )
    @pytest.mark.closes_loop
    def test_cli(self, cli_entry_point, monkeypatch, request, args, cancel_after_timeout, event_loop):
        args = list(itertools.chain.from_iterable(arg.split(' ') for arg in args))
        monkeypatch.setattr('sys.argv',
                            [os.fspath(request.path)] + args,
                            raising=True)
        import sys
        assert sys.argv[1:] == args
        cli_entry_point()

        from ubii.cli.main import parse_args
        from ubii.framework import client

        used_client: client.UbiiClient = list(client.UbiiClient.registry.values())[-1]
        assert used_client.implements(client.InitProcessingModules)

        args = parse_args()
        for name, mod in args.processing_modules:
            modargs = {n:v for n, v in args.module_args}.get(name, {})
            instance = ubii.framework.processing.ProcessingRoutine.registry.get(name)

            for arg, value in modargs.items():
                assert getattr(instance, arg) == value



from __future__ import annotations

import asyncio
import logging
import typing

from ubii.framework import (
    client,
    constants,
    protocol,
)
from ubii.framework.util.typing import Protocol
from . import protocol as default_protocol_

P = typing.TypeVar('P', bound=protocol.AbstractProtocol)

log = logging.getLogger(__name__)


class connect(typing.Awaitable[client.UbiiClient[P]],
              typing.AsyncContextManager[client.UbiiClient[P]],
              typing.ContextManager[client.UbiiClient[P]],
              typing.Generic[P]):
    """
    Use this callable to easily get a running Ubi Interact Node. If you don't want to change the
    node behaviour significantly, this is the go-to to connect with the `master node`.


    Example:

        Use the default :class:`connect` callable with defaults to connect to a `master node`.
        If :class:`connect` uses a ``url`` argument, it takes precedence over the
        :attr:`~ubii.framework.constants.UbiiConfig.DEFAULT_SERVICE_URL` of the ``config`` argument.
        If :class:`connect` does not use a specific constants, the shared
        :attr:`~ubii.framework.constants.GLOBAL_CONFIG` is used by default.

        There are multiple ways to create the client with :class:`connect`.

            with :ref:`await` ::

                from ubii.node.connect import connect
                import asyncio

                async def main():
                    global client
                    client = await connect()
                    assert client.id

                    # do something with the client
                    ...

                asyncio.run(main())


    """
    class ClientFactory(Protocol):
        def __call__(self,
                     instance: connect,
                     *,
                     client_type: typing.Type[client.UbiiClient],
                     protocol_type: typing.Type[P]) -> client.UbiiClient[P]:
            """
            :class:`ClientFactory` objects need to have this call signature

            Args:
                instance: a :class:`connect` instance
                client_type: a :class:`~ubii.framework.client.UbiiClient` type
                protocol_type: a :class:`~ubii.framework.protocol.AbstractProtocol` type

            Returns:
                A client using the protocol
            """

    def __init__(self: connect[P],
                 url=None,
                 config: constants.UbiiConfig = constants.GLOBAL_CONFIG,
                 client_type: typing.Type[client.UbiiClient] = client.UbiiClient,
                 protocol_type: typing.Type[P] = default_protocol_.DefaultProtocol):
        """
        Args:
            url: URL of the `master node`, overwrites the
                :attr:`.ubii.framework.constants.UbiiConfig.DEFAULT_SERVICE_URL` of the used ``config``
            config: contains information like default topic for `server configuration` service call
            client_type: a :class:`~ubii.framework.client.UbiiClient` type
            protocol_type: a :class:`~ubii.framework.protocol.AbstractProtocol` type
        """
        if url is not None:
            config.DEFAULT_SERVICE_URL = url

        self.config: constants.UbiiConfig = config
        """
        reference to used config
        """

        factory = self.client_factories.get((client_type, protocol_type))
        if not factory:
            raise ValueError(f"No client factory found for {(client_type, protocol_type)}")
        else:
            log.debug(f"Using {factory} to create {client_type} with {protocol_type}")

        self.client: client.UbiiClient[P] = factory(
            self,
            client_type=client_type,
            protocol_type=protocol_type
        )
        """
        Will be created with a callable from :attr:`.client_factories` if it contains one for the 
        ``client_type`` and ``protocol_type``
        """

    def default_create(self, *, client_type, protocol_type):
        """
        Simply instantiates the ``protocol_type`` (passing :attr:`.config` as ``config``),
        creates a client with this protocol, and sets the protocols
        :attr:`~ubii.framework.client.AbstractClientProtocol.client` attribute

        Args:
            client_type:
            protocol_type:

        Returns:
            the client using the protocol
        """
        protocol = protocol_type(config=self.config)
        client = client_type(protocol=protocol)
        protocol.client = client
        return client

    def __await__(self) -> typing.Generator[typing.Any, None, client.UbiiClient[P]]:
        return self.client.__await__()

    def __aenter__(self):
        return self.client.__aenter__()

    def __aexit__(self, *exc_infos):
        return self.client.__aexit__(*exc_infos)

    def __enter__(self: connect[P]) -> client.UbiiClient[P]:
        return self.client

    def __exit__(self, *exc_info):
        self.client.protocol.task_nursery.create_task(self.client.protocol.stop())

    client_factories: typing.Dict[
        typing.Tuple[typing.Type[client.UbiiClient], typing.Type[protocol.AbstractProtocol]], ClientFactory] = {
        (client.UbiiClient, default_protocol_.DefaultProtocol): default_create
    }



"""
Global variables are problematic, but global constants are a useful feature especially as fallback for
configuration options.

Warning:

    Keep in mind that assignments to global variables are done when the module is first imported so to make sure
    to have the "correct" version of the constants defined in this module, import the module, not
    the individual names

    Good:
        >>> from ubii.framework import constants
        >>> config = constants.GLOBAL_CONFIG

    Bad:
        >>> from ubii.framework.constants import GLOBAL_GONFIG
"""

from __future__ import annotations

import dataclasses
import os
import copy

import ubii.proto

__protobuf__ = ubii.proto.__protobuf__

UBII_URL_ENV = 'UBII_SERVICE_URL'
"""
Setting this environment variable affects the :attr:`default_service_url`
"""

default_constants: ubii.proto.Constants = ubii.proto.Constants()
"""
Constants with default topic for `server configuration` service set (needed to get all other default topics from 
`master node`)
"""
default_constants.DEFAULT_TOPICS.SERVICES.SERVER_CONFIG = '/services/server_configuration'
default_server: ubii.proto.Server = ubii.proto.Server()
"""
Server with :attr:`~ubii.proto.Server.constants_json` set to :attr:`default_contants` (converted to JSON)
"""
default_server.constants_json = ubii.proto.Constants.to_json(default_constants)
default_service_url = os.getenv(UBII_URL_ENV, 'http://localhost:8102/services/json')
"""
Global variable for url of `master node` service backend. Read from :attr:`UBII_URL_ENV` environment variable, 
fallback ``http://localhost:8102/services/json``

:meta hide-value:
"""


@dataclasses.dataclass
class UbiiConfig:
    """
    Config options for the Ubi interact node.

    Note:
        Uses *copies* of global attributes from this module as default

            *   :attr:`.SERVER` = :attr:`default_server`
            *   :attr:`.CONSTANTS` = :attr:`default_constants`
            *   :attr:`.DEFAULT_SERVICE_URL` = :attr:`default_service_url`

        i.e. assigning to those values does not change the global defaults from this module

        >>> from ubii.framework import constants
        >>> config = constants.UbiiConfig()
        >>> config.SERVER == constants.default_server
        True
        >>> config.SERVER.constants_json = "Foo"
        >>> config.SERVER == constants.default_server
        False

    """

    SERVER: ubii.proto.Server = dataclasses.field(default_factory=lambda: copy.copy(default_server))
    """
    needed for all service calls, and typically provided by the master node.
    To get the config the defaults include the topic for the `server configuration` service.
    """
    CONSTANTS: ubii.proto.Constants = dataclasses.field(default_factory=lambda: copy.copy(default_constants))
    """
    includes all meta information about the master node (ip address, ports, etc.)
    Currently the :attr:`ubii.proto.Server.constants_json` field should be parsed as a :class:`ubii.proto.Constants` 
    message and updated in your config whenever the Server is updated (at some point the master node might start sending
    actual proto messages instead of just JSON)
    """
    DEFAULT_SERVICE_URL: str = default_service_url
    """
    needed to make the first service request (`server configuration`) before anything
    else is known (see also :attr:`UBII_URL_ENV`)
    
    :meta hide-value:
    """


GLOBAL_CONFIG = UbiiConfig()
"""
This is a global :class:`UbiiConfig` object, typically used as default configuration for client protocols.
Use this global config to share configuration changes to values in a global scope.

:meta hide-value:
"""

__all__ = [
    "GLOBAL_CONFIG",
    "UBII_URL_ENV",
    "UbiiConfig",
    "default_service_url",
    "default_server",
    "default_constants"
]



"""
You may import common objects directly from here
"""

# Make sure to import all modules with import side effects (e.g. marshal rule updates)
from .constants import (
    UbiiConfig,
    GLOBAL_CONFIG
)
from .errors import (
    UbiiError,
    SessionRuntimeStopServiceError
)
from .logging import (
    logging_setup,
    parse_args,
)
from .util import debug

from .client import (
    UbiiClient,
)

__all__ = (
    'GLOBAL_CONFIG',
    'UbiiConfig',
    'UbiiError',
    'UbiiClient',
    'SessionRuntimeStopServiceError',
    'logging_setup',
    'parse_args',
    'debug'
)



import asyncio

import pytest
import ubii.proto as ub

from ubii.framework import debug

__protobuf__ = ub.__protobuf__


class TestBasic:
    async def test_debug_settings(self, debug_settings):
        assert debug()


class TestConnection:
    async def connection_context_manager(self, connect):
        async with connect() as connected:
            client = connected
            await asyncio.sleep(4)
            return client

    async def connection_await(self, connect):
        client = await connect()
        await asyncio.sleep(4)
        return client

    @pytest.mark.parametrize('connection', [connection_await, connection_context_manager])
    async def test_connect_client(self, connection):
        from ubii.node import connect_client
        from ubii.framework.client import UbiiClient

        client: UbiiClient = await connection.__get__(self, type(self))(connect_client)
        assert client.id
        assert client.protocol.client == client

        # when using the context manager, the protocol is already stopped
        protocol_expected_ended = connection == type(self).connection_context_manager
        assert (client.protocol.state.value == client.protocol.end_state) == protocol_expected_ended
